from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import ClientMessage
from hazelcast.protocol.codec.ringbuffer_message_type import *

REQUEST_TYPE = RINGBUFFER_REMAININGCAPACITY
RESPONSE_TYPE = 103
RETRYABLE = True


def calculate_size(name):
    """ Calculates the request payload size"""
    data_size = 0
    data_size += calculate_size_str(name)
    return data_size


def encode_request(name):
    """ Encode request into client_message"""
    client_message = ClientMessage(payload_size=calculate_size(name))
    client_message.set_message_type(REQUEST_TYPE)
    client_message.set_retryable(RETRYABLE)
    client_message.append_str(name)
    client_message.update_frame_length()
    return client_message


def decode_response(client_message, to_object=None):
    """ Decode response from client message"""
    parameters = dict(response=None)
    parameters['response'] = client_message.read_long()
    return parameters



from hazelcast.protocol.codec import \
    semaphore_acquire_codec, \
    semaphore_available_permits_codec, \
    semaphore_drain_permits_codec, \
    semaphore_init_codec, \
    semaphore_reduce_permits_codec, \
    semaphore_release_codec, \
    semaphore_try_acquire_codec
from hazelcast.proxy.base import PartitionSpecificProxy
from hazelcast.util import check_not_negative, to_millis


class Semaphore(PartitionSpecificProxy):
    """
    Semaphore is a backed-up distributed alternative to the Python `asyncio.Semaphore <https://docs.python.org/3/library/asyncio-sync.html>`_

    Semaphore is a cluster-wide counting semaphore. Conceptually, it maintains a set of permits. Each acquire() blocks
    if necessary until a permit is available, and then takes it. Each release() adds a permit, potentially releasing a
    blocking acquirer. However, no actual permit objects are used; the semaphore just keeps a count of the number
    available and acts accordingly.

    The Hazelcast distributed semaphore implementation guarantees that threads invoking any of the acquire methods are
    selected to obtain permits in the order in which their invocation of those methods was processed(first-in-first-out;
    FIFO). Note that FIFO ordering necessarily applies to specific internal points of execution within the cluster.
    Therefore, it is possible for one member to invoke acquire before another, but reach the ordering point after the
    other, and similarly upon return from the method.

    This class also provides convenience methods to acquire and release multiple permits at a time. Beware of the
    increased risk of indefinite postponement when using the multiple acquire. If a single permit is released to a
    semaphore that is currently blocking, a thread waiting for one permit will acquire it before a thread waiting for
    multiple permits regardless of the call order.
    """
    def init(self, permits):
        """
        Try to initialize this Semaphore instance with the given permit count.

        :param permits: (int), the given permit count.
        :return: (bool), ``true`` if initialization success.
        """
        check_not_negative(permits, "Permits cannot be negative!")
        return self._encode_invoke(semaphore_init_codec, permits=permits)

    def acquire(self, permits=1):
        """
        Acquires one or specified amount of permits if available, and returns immediately, reducing the number of
        available permits by one or given amount.

        If insufficient permits are available then the current thread becomes disabled for thread scheduling purposes
        and lies dormant until one of following happens:

            * some other thread invokes one of the release methods for this semaphore, the current thread is next to be
            assigned permits and the number of available permits satisfies this request,
            * this Semaphore instance is destroyed, or
            * some other thread interrupts the current thread.

        :param permits: (int), the number of permits to acquire (optional).
        """
        check_not_negative(permits, "Permits cannot be negative!")
        return self._encode_invoke(semaphore_acquire_codec, permits=permits)

    def available_permits(self):
        """
        Returns the current number of permits currently available in this semaphore.

            * This method is typically used for debugging and testing purposes.
        :return: (int), the number of available permits in this semaphore.
        """
        return self._encode_invoke(semaphore_available_permits_codec)

    def drain_permits(self):
        """
        Acquires and returns all permits that are immediately available.

        :return: (int), the number of permits drained.
        """
        return self._encode_invoke(semaphore_drain_permits_codec)

    def reduce_permits(self, reduction):
        """
        Shrinks the number of available permits by the indicated reduction. This method differs from acquire in that it
        does not block waiting for permits to become available.

        :param reduction: (int),  the number of permits to remove.
        """
        check_not_negative(reduction, "Reduction cannot be negative!")
        return self._encode_invoke(semaphore_reduce_permits_codec, reduction=reduction)

    def release(self, permits=1):
        """
        Releases one or given number of permits, increasing the number of available permits by one or that amount.

        There is no requirement that a thread that releases a permit must have acquired that permit by calling one of
        the acquire methods. Correct usage of a semaphore is established by programming convention in the application.

        :param permits: (int), the number of permits to release (optional).
        """
        check_not_negative(permits, "Permits cannot be negative!")
        return self._encode_invoke(semaphore_release_codec, permits=permits)

    def try_acquire(self, permits=1, timeout=0):
        """
        Tries to acquire one or the given number of permits, if they are available, and returns immediately, with the
        value ``true``, reducing the number of available permits by the given amount.

        If there are insufficient permits and a timeout is provided, the current thread becomes disabled for thread
        scheduling purposes and lies dormant until one of following happens:
            * some other thread invokes the release() method for this semaphore and the current thread is next to be
            assigned a permit, or
            * some other thread interrupts the current thread, or
            * the specified waiting time elapses.

        If there are insufficient permits and no timeout is provided, this method will return immediately with the value
        ``false`` and the number of available permits is unchanged.

        :param permits: (int), the number of permits to acquire (optional).
        :param timeout: (long), the maximum time in seconds to wait for the permit(s) (optional).
        :return: (bool), ``true`` if desired amount of permits was acquired, ``false`` otherwise.
        """
        check_not_negative(permits, "Permits cannot be negative!")
        return self._encode_invoke(semaphore_try_acquire_codec, permits=permits, timeout=to_millis(timeout))



from hazelcast.serialization.api import IdentifiedDataSerializable

PREDICATE_FACTORY_ID = -32


class Predicate(IdentifiedDataSerializable):
    def get_factory_id(self):
        return PREDICATE_FACTORY_ID

    def get_class_id(self):
        return self.CLASS_ID


class SqlPredicate(Predicate):
    CLASS_ID = 0

    def __init__(self, sql=None):
        self.sql = sql

    def write_data(self, output):
        output.write_utf(self.sql)

    def __repr__(self):
        return "SqlPredicate(sql='%s')" % self.sql


class AndPredicate(Predicate):
    CLASS_ID = 1

    def __init__(self, *predicates):
        self.predicates = predicates

    def write_data(self, output):
        output.write_int(len(self.predicates))
        for predicate in self.predicates:
            output.write_object(predicate)

    def __repr__(self):
        return "AndPredicate(%s)" % ", ".join([str(p) for p in self.predicates])


class BetweenPredicate(Predicate):
    CLASS_ID = 2

    def __init__(self, attribute, from_, to):
        self.attribute = attribute
        self.from_ = from_
        self.to = to

    def write_data(self, output):
        output.write_utf(self.attribute)
        output.write_object(self.to)
        output.write_object(self.from_)

    def __repr__(self):
        return "BetweenPredicate(attribute='%s', from=%s, to=%s)" % (self.attribute, self.from_, self.to)


class EqualPredicate(Predicate):
    CLASS_ID = 3

    def __init__(self, attribute, value):
        self.attribute = attribute
        self.value = value

    def write_data(self, output):
        output.write_utf(self.attribute)
        output.write_object(self.value)

    def __repr__(self):
        return "EqualPredicate(attribute='%s', value=%s)" % (self.attribute, self.value)


class GreaterLessPredicate(Predicate):
    CLASS_ID = 4

    def __init__(self, attribute, value, is_equal, is_less):
        self.attribute = attribute
        self.value = value
        self.is_equal = is_equal
        self.is_less = is_less

    def write_data(self, output):
        output.write_utf(self.attribute)
        output.write_object(self.value)
        output.write_boolean(self.is_equal)
        output.write_boolean(self.is_less)

    def __repr__(self):
        return "GreaterLessPredicate(attribute='%s', value=%s, is_equal=%s, is_less=%s)" % (
            self.attribute, self.value, self.is_equal, self.is_less)


class LikePredicate(Predicate):
    CLASS_ID = 5

    def __init__(self, attribute, expression):
        self.attribute = attribute
        self.expression = expression

    def write_data(self, output):
        output.write_utf(self.attribute)
        output.write_utf(self.expression)

    def __repr__(self):
        return "LikePredicate(attribute='%s', expression='%s')" % (self.attribute, self.expression)


class ILikePredicate(LikePredicate):
    CLASS_ID = 6

    def __repr__(self):
        return "ILikePredicate(attribute='%s', expression='%s')" % (self.attribute, self.expression)


class InPredicate(Predicate):
    CLASS_ID = 7

    def __init__(self, attribute, *values):
        self.attribute = attribute
        self.values = values

    def write_data(self, output):
        output.write_utf(self.attribute)
        output.write_int(len(self.values))
        for value in self.values:
            output.write_object(value)

    def __repr__(self):
        return "InPredicate(attribute='%s', %s)" % (self.attribute, ",".join([str(x) for x in self.values]))


class InstanceOfPredicate(Predicate):
    CLASS_ID = 8

    def __init__(self, class_name):
        self.class_name = class_name

    def write_data(self, output):
        output.write_utf(self.class_name)

    def __repr__(self):
        return "InstanceOfPredicate(class_name='%s')" % self.class_name


class NotEqualPredicate(EqualPredicate):
    CLASS_ID = 9

    def __repr__(self):
        return "NotEqualPredicate(attribute='%s', value=%s)" % (self.attribute, self.value)


class NotPredicate(Predicate):
    CLASS_ID = 10

    def __init__(self, predicate):
        self.predicate = predicate

    def write_data(self, output):
        output.write_object(self.predicate)

    def __repr__(self):
        return "NotPredicate(predicate=%s)" % self.predicate


class OrPredicate(AndPredicate):
    CLASS_ID = 11

    def __repr__(self):
        return "OrPredicate(%s)" % ", ".join([str(p) for p in self.predicates])


class RegexPredicate(Predicate):
    CLASS_ID = 12

    def __init__(self, attribute, pattern):
        self.attribute = attribute
        self.pattern = pattern

    def write_data(self, output):
        output.write_utf(self.attribute)
        output.write_utf(self.pattern)

    def __repr__(self):
        return "RegexPredicate(attribute='%s', pattern='%s')" % (self.attribute, self.pattern)


class FalsePredicate(Predicate):
    CLASS_ID = 13

    def write_data(self, object_data_output):
        pass

    def __repr__(self):
        return "FalsePredicate()"


class TruePredicate(Predicate):
    CLASS_ID = 14

    def write_data(self, object_data_output):
        pass

    def __repr__(self):
        return "TruePredicate()"

sql = SqlPredicate
is_equal_to = EqualPredicate
is_not_equal_to = NotEqualPredicate
is_like = LikePredicate
is_ilike = ILikePredicate
matches_regex = RegexPredicate
and_ = AndPredicate
or_ = OrPredicate
not_ = NotPredicate
is_between = BetweenPredicate
is_in = InPredicate
is_instance_of = InstanceOfPredicate
is_not = NotPredicate
false = FalsePredicate
true = TruePredicate


def is_greater_than(attribute, x):
    return GreaterLessPredicate(attribute, x, False, False)


def is_greater_than_or_equal_to(attribute, x):
    return GreaterLessPredicate(attribute, x, True, False)


def is_less_than(attribute, x):
    return GreaterLessPredicate(attribute, x, False, True)


def is_less_than_or_equal_to(attribute, x):
    return GreaterLessPredicate(attribute, x, True, True)





REPLICATEDMAP_PUT = 0x0e01
REPLICATEDMAP_SIZE = 0x0e02
REPLICATEDMAP_ISEMPTY = 0x0e03
REPLICATEDMAP_CONTAINSKEY = 0x0e04
REPLICATEDMAP_CONTAINSVALUE = 0x0e05
REPLICATEDMAP_GET = 0x0e06
REPLICATEDMAP_REMOVE = 0x0e07
REPLICATEDMAP_PUTALL = 0x0e08
REPLICATEDMAP_CLEAR = 0x0e09
REPLICATEDMAP_ADDENTRYLISTENERTOKEYWITHPREDICATE = 0x0e0a
REPLICATEDMAP_ADDENTRYLISTENERWITHPREDICATE = 0x0e0b
REPLICATEDMAP_ADDENTRYLISTENERTOKEY = 0x0e0c
REPLICATEDMAP_ADDENTRYLISTENER = 0x0e0d
REPLICATEDMAP_REMOVEENTRYLISTENER = 0x0e0e
REPLICATEDMAP_KEYSET = 0x0e0f
REPLICATEDMAP_VALUES = 0x0e10
REPLICATEDMAP_ENTRYSET = 0x0e11
REPLICATEDMAP_ADDNEARCACHEENTRYLISTENER = 0x0e12




MULTIMAP_PUT = 0x0201
MULTIMAP_GET = 0x0202
MULTIMAP_REMOVE = 0x0203
MULTIMAP_KEYSET = 0x0204
MULTIMAP_VALUES = 0x0205
MULTIMAP_ENTRYSET = 0x0206
MULTIMAP_CONTAINSKEY = 0x0207
MULTIMAP_CONTAINSVALUE = 0x0208
MULTIMAP_CONTAINSENTRY = 0x0209
MULTIMAP_SIZE = 0x020a
MULTIMAP_CLEAR = 0x020b
MULTIMAP_VALUECOUNT = 0x020c
MULTIMAP_ADDENTRYLISTENERTOKEY = 0x020d
MULTIMAP_ADDENTRYLISTENER = 0x020e
MULTIMAP_REMOVEENTRYLISTENER = 0x020f
MULTIMAP_LOCK = 0x0210
MULTIMAP_TRYLOCK = 0x0211
MULTIMAP_ISLOCKED = 0x0212
MULTIMAP_UNLOCK = 0x0213
MULTIMAP_FORCEUNLOCK = 0x0214
MULTIMAP_REMOVEENTRY = 0x0215



"""
Hazelcast client protocol codecs
"""

from collections import namedtuple
from hazelcast.core import Member, DistributedObjectInfo, EntryView, Address
from hazelcast.six.moves import range

EXCEPTION_MESSAGE_TYPE = 109


class MemberCodec(object):
    @classmethod
    def encode(cls, client_message, member):
        AddressCodec.encode(client_message, member.address)
        client_message.append_str(member.uuid)
        client_message.append_bool(member.is_lite_member)
        client_message.append_int(len(member.attributes))
        for key, value in member.attributes:
            client_message.append_str(key)
            client_message.append_str(value)

    @classmethod
    def decode(cls, client_message, to_object=None):
        address = AddressCodec.decode(client_message)
        uuid = client_message.read_str()
        lite_member = client_message.read_bool()
        attribute_size = client_message.read_int()
        attributes = {}
        for i in range(0, attribute_size, 1):
            key = client_message.read_str()
            value = client_message.read_str()
            attributes[key] = value
        return Member(address, uuid, lite_member, attributes)


class AddressCodec(object):
    @classmethod
    def encode(cls, client_message, obj):
        client_message.append_str(obj.host).append_int(obj.port)

    @classmethod
    def decode(cls, client_message, to_object=None):
        host = client_message.read_str()
        port = client_message.read_int()
        return Address(host, port)


class DistributedObjectInfoCodec(object):
    @classmethod
    def encode(cls, client_message, obj):
        client_message.append_str(obj.service_name).append_str(obj.name)

    @classmethod
    def decode(cls, client_message):
        service_name = client_message.read_str()
        name = client_message.read_str()
        return DistributedObjectInfo(name, service_name)


class EntryViewCodec(object):
    @classmethod
    def encode(cls, client_message, entry_view):
        client_message.append_data(entry_view.key)
        client_message.append_data(entry_view.value)
        client_message.append_long(entry_view.cost)
        client_message.append_long(entry_view.cost)
        client_message.append_long(entry_view.creationTime)
        client_message.append_long(entry_view.expirationTime)
        client_message.append_long(entry_view.hits)
        client_message.append_long(entry_view.lastAccessTime)
        client_message.append_long(entry_view.lastStoredTime)
        client_message.append_long(entry_view.lastUpdateTime)
        client_message.append_long(entry_view.version)
        client_message.append_long(entry_view.evictionCriteriaNumber)
        client_message.append_long(entry_view.ttl)

    @classmethod
    def decode(cls, client_message, to_object):
        entry_view = EntryView()
        entry_view.key = to_object(client_message.read_data())
        entry_view.value = to_object(client_message.read_data())
        entry_view.cost = client_message.read_long()
        entry_view.creation_time = client_message.read_long()
        entry_view.expiration_time = client_message.read_long()
        entry_view.hits = client_message.read_long()
        entry_view.last_access_time = client_message.read_long()
        entry_view.last_stored_time = client_message.read_long()
        entry_view.last_update_time = client_message.read_long()
        entry_view.version = client_message.read_long()
        entry_view.eviction_criteria_number = client_message.read_long()
        entry_view.ttl = client_message.read_long()
        return entry_view


class QueryCacheEventDataCodec(object):
    @classmethod
    def encode(cls, client_message, obj):
        pass

    @classmethod
    def decode(cls, client_message, to_object=None):
        pass


StackTraceElement = namedtuple('StackTraceElement', ['declaring_class', 'method_name', 'file_name', 'line_number'])


class ErrorCodec(object):
    message = None
    cause_class_name = None

    def __init__(self, client_message):
        self.error_code = client_message.read_int()
        self.class_name = client_message.read_str()
        if not client_message.read_bool():
            self.message = client_message.read_str()

        self.stack_trace = []
        stack_trace_count = client_message.read_int()
        for _ in range(stack_trace_count):
            self.stack_trace.append(self.decode_stack_trace(client_message))

        self.cause_error_code = client_message.read_int()
        if not client_message.read_bool():
            self.cause_class_name = client_message.read_str()

    @staticmethod
    def decode_stack_trace(client_message):
        declaring_class = client_message.read_str()
        method_name = client_message.read_str()
        file_name = None
        if not client_message.read_bool():
            file_name = client_message.read_str()
        line_number = client_message.read_int()
        return StackTraceElement(declaring_class=declaring_class,
                                 method_name=method_name, file_name=file_name, line_number=line_number)

    def __repr__(self):
        return 'ErrorCodec(error_code="%s", class_name="%s", message="%s", cause_error_code="%s", ' \
               'cause_class_name="%s' % (self.error_code, self.class_name, self.message, self.cause_error_code,
                                         self.cause_class_name)



import logging
import threading
import time
import functools

from hazelcast.exception import create_exception, HazelcastInstanceNotActiveError, is_retryable_error, TimeoutError, \
    TargetDisconnectedError, HazelcastClientNotActiveException, TargetNotMemberError
from hazelcast.future import Future
from hazelcast.lifecycle import LIFECYCLE_STATE_CONNECTED
from hazelcast.protocol.client_message import LISTENER_FLAG
from hazelcast.protocol.custom_codec import EXCEPTION_MESSAGE_TYPE, ErrorCodec
from hazelcast.util import AtomicInteger
from hazelcast.six.moves import queue
from hazelcast import six


class Invocation(object):
    sent_connection = None
    timer = None

    def __init__(self, invocation_service, request, partition_id=-1, address=None, connection=None, event_handler=None):
        self._event = threading.Event()
        self._invocation_timeout = invocation_service.invocation_timeout
        self.timeout = self._invocation_timeout + time.time()
        self.address = address
        self.connection = connection
        self.partition_id = partition_id
        self.request = request
        self.future = Future()
        self.event_handler = event_handler

    def has_connection(self):
        return self.connection is not None

    def has_partition_id(self):
        return self.partition_id >= 0

    def has_address(self):
        return self.address is not None

    def set_response(self, response):
        if self.timer:
            self.timer.cancel()
        self.future.set_result(response)

    def set_exception(self, exception, traceback=None):
        if self.timer:
            self.timer.cancel()
        self.future.set_exception(exception, traceback)

    def set_timeout(self, timeout):
        self._invocation_timeout = timeout
        self.timeout = self._invocation_timeout + time.time()

    def on_timeout(self):
        self.set_exception(TimeoutError("Request timed out after %d seconds." % self._invocation_timeout))


class InvocationService(object):
    logger = logging.getLogger("HazelcastClient.InvocationService")

    def __init__(self, client):
        self._pending = {}
        self._next_correlation_id = AtomicInteger(1)
        self._client = client
        self._logger_extras = {"client_name": client.name, "group_name": client.config.group_config.name}
        self._event_queue = queue.Queue()
        self._is_redo_operation = client.config.network_config.redo_operation
        self.invocation_retry_pause = self._init_invocation_retry_pause()
        self.invocation_timeout = self._init_invocation_timeout()
        self._listener_service = None

        if client.config.network_config.smart_routing:
            self.invoke = self.invoke_smart
        else:
            self.invoke = self.invoke_non_smart

        self._client.connection_manager.add_listener(on_connection_closed=self.cleanup_connection)
        client.heartbeat.add_listener(on_heartbeat_stopped=self._heartbeat_stopped)

    def start(self):
        self._listener_service = self._client.listener

    def invoke_on_connection(self, message, connection, ignore_heartbeat=False, event_handler=None):
        return self.invoke(Invocation(self, message, connection=connection, event_handler=event_handler),
                           ignore_heartbeat)

    def invoke_on_partition(self, message, partition_id, invocation_timeout=None):
        invocation = Invocation(self, message, partition_id=partition_id)
        if invocation_timeout:
            invocation.set_timeout(invocation_timeout)
        return self.invoke(invocation)

    def invoke_on_random_target(self, message):
        return self.invoke(Invocation(self, message))

    def invoke_on_target(self, message, address):
        return self.invoke(Invocation(self, message, address=address))

    def invoke_smart(self, invocation, ignore_heartbeat=False):
        if invocation.has_connection():
            self._send(invocation, invocation.connection, ignore_heartbeat)
        elif invocation.has_partition_id():
            addr = self._client.partition_service.get_partition_owner(invocation.partition_id)
            if addr is None:
                self._handle_exception(invocation, IOError("Partition does not have an owner. "
                                                           "partition Id: ".format(invocation.partition_id)))
            elif not self._is_member(addr):
                self._handle_exception(invocation, TargetNotMemberError("Partition owner '{}' "
                                                                        "is not a member.".format(addr)))
            else:
                self._send_to_address(invocation, addr)
        elif invocation.has_address():
            if not self._is_member(invocation.address):
                self._handle_exception(invocation, TargetNotMemberError("Target '{}' is not a member.".format
                                                                        (invocation.address)))
            else:
                self._send_to_address(invocation, invocation.address)
        else:  # send to random address
            addr = self._client.load_balancer.next_address()
            if addr is None:
                self._handle_exception(invocation, IOError("No address found to invoke"))
            else:
                self._send_to_address(invocation, addr)
        return invocation.future

    def invoke_non_smart(self, invocation, ignore_heartbeat=False):
        if invocation.has_connection():
            self._send(invocation, invocation.connection, ignore_heartbeat)
        else:
            addr = self._client.cluster.owner_connection_address
            self._send_to_address(invocation, addr)
        return invocation.future

    def cleanup_connection(self, connection, cause):
        for correlation_id, invocation in six.iteritems(dict(self._pending)):
            if invocation.sent_connection == connection:
                self._handle_exception(invocation, cause)

    def _init_invocation_retry_pause(self):
        invocation_retry_pause = self._client.properties.get_seconds_positive_or_default(
            self._client.properties.INVOCATION_RETRY_PAUSE_MILLIS)
        return invocation_retry_pause

    def _init_invocation_timeout(self):
        invocation_timeout = self._client.properties.get_seconds_positive_or_default(
            self._client.properties.INVOCATION_TIMEOUT_SECONDS)
        return invocation_timeout

    def _heartbeat_stopped(self, connection):
        for correlation_id, invocation in six.iteritems(dict(self._pending)):
            if invocation.sent_connection == connection:
                self._handle_exception(invocation,
                                       TargetDisconnectedError("%s has stopped heart beating." % connection))

    def _send_to_address(self, invocation, address, ignore_heartbeat=False):
        try:
            conn = self._client.connection_manager.connections[address]
            self._send(invocation, conn, ignore_heartbeat)
        except KeyError:
            if self._client.lifecycle.state != LIFECYCLE_STATE_CONNECTED:
                self._handle_exception(invocation, IOError("Client is not in connected state"))
            else:
                self._client.connection_manager.get_or_connect(address).continue_with(self.on_connect, invocation,
                                                                                      ignore_heartbeat)

    def on_connect(self, f, invocation, ignore_heartbeat):
        if f.is_success():
            self._send(invocation, f.result(), ignore_heartbeat)
        else:
            self._handle_exception(invocation, f.exception(), f.traceback())

    def _send(self, invocation, connection, ignore_heartbeat):
        correlation_id = self._next_correlation_id.get_and_increment()
        message = invocation.request
        message.set_correlation_id(correlation_id)
        message.set_partition_id(invocation.partition_id)
        self._pending[correlation_id] = invocation
        if not invocation.timer:
            invocation.timer = self._client.reactor.add_timer_absolute(invocation.timeout, invocation.on_timeout)

        if invocation.event_handler is not None:
            self._listener_service.add_event_handler(correlation_id, invocation.event_handler)

        self.logger.debug("Sending %s to %s", message, connection, extra=self._logger_extras)

        if not ignore_heartbeat and not connection.heartbeating:
            self._handle_exception(invocation, TargetDisconnectedError("%s has stopped heart beating." % connection))
            return

        invocation.sent_connection = connection
        try:
            connection.send_message(message)
        except IOError as e:
            if invocation.event_handler is not None:
                self._listener_service.remove_event_handler(correlation_id)
            self._handle_exception(invocation, e)

    def _handle_client_message(self, message):
        correlation_id = message.get_correlation_id()
        if message.has_flags(LISTENER_FLAG):
            self._listener_service.handle_client_message(message)
            return
        if correlation_id not in self._pending:
            self.logger.warning("Got message with unknown correlation id: %s", message, extra=self._logger_extras)
            return
        invocation = self._pending.pop(correlation_id)

        if message.get_message_type() == EXCEPTION_MESSAGE_TYPE:
            error = create_exception(ErrorCodec(message))
            return self._handle_exception(invocation, error)

        invocation.set_response(message)

    def _handle_event(self, invocation, message):
        try:
            invocation.event_handler(message)
        except:
            self.logger.warning("Error handling event %s", message, exc_info=True, extra=self._logger_extras)

    def _handle_exception(self, invocation, error, traceback=None):
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug("Got exception for request %s: %s: %s", invocation.request, type(error).__name__, error,
                              extra=self._logger_extras)

        if not self._client.lifecycle.is_live:
            invocation.set_exception(HazelcastClientNotActiveException(error.args[0]), traceback)
            return

        if self._is_not_allowed_to_retry_on_selection(invocation, error):
            invocation.set_exception(error, traceback)
            return

        if not self._should_retry(invocation, error):
            invocation.set_exception(error, traceback)
            return

        if invocation.timeout < time.time():
            if self.logger.isEnabledFor(logging.DEBUG):
                self.logger.debug('Error will not be retried because invocation timed out: %s', error,
                                  extra=self._logger_extras)
            invocation.set_excetion(TimeoutError(
                '%s timed out because an error occurred after invocation timeout: %s' % (invocation.request, error),
                traceback))
            return

        invoke_func = functools.partial(self.invoke, invocation)
        self._client.reactor.add_timer(self.invocation_retry_pause, invoke_func)

    def _should_retry(self, invocation, error):
        if isinstance(error, (IOError, HazelcastInstanceNotActiveError)) or is_retryable_error(error):
            return True

        if isinstance(error, TargetDisconnectedError):
            return invocation.request.is_retryable() or self._is_redo_operation

        return False

    def _is_not_allowed_to_retry_on_selection(self, invocation, error):
        if invocation.connection is not None and isinstance(error, IOError):
            return True

        # When invocation is sent over an address,error is the TargetNotMemberError and the
        # member is not in the member list, we should not retry
        return invocation.address is not None and isinstance(error, TargetNotMemberError) \
               and not self._is_member(invocation.address)

    def _is_member(self, address):
        return self._client.cluster.get_member_by_address(address) is not None



from hazelcast.protocol.codec import transactional_queue_offer_codec, transactional_queue_peek_codec, \
    transactional_queue_poll_codec, transactional_queue_size_codec, transactional_queue_take_codec
from hazelcast.proxy.base import TransactionalProxy
from hazelcast.util import check_not_none, to_millis


class TransactionalQueue(TransactionalProxy):
    """
    Transactional implementation of :class:`~hazelcast.proxy.queue.Queue`.
    """
    def offer(self, item, timeout=0):
        """
        Transactional implementation of :func:`Queue.offer(item, timeout) <hazelcast.proxy.queue.Queue.offer>`

        :param item: (object), the item to be added.
        :param timeout: (long), maximum time in seconds to wait for addition (optional).
        :return: (bool), ``true`` if the element was added to this queue, ``false`` otherwise.
        """
        check_not_none(item, "item can't be none")
        return self._encode_invoke(transactional_queue_offer_codec, item=self._to_data(item),
                                   timeout=to_millis(timeout))

    def take(self):
        """
        Transactional implementation of :func:`Queue.take() <hazelcast.proxy.queue.Queue.take>`

        :return: (object), the head of this queue.
        """
        return self._encode_invoke(transactional_queue_take_codec)

    def poll(self, timeout=0):
        """
        Transactional implementation of :func:`Queue.poll(timeout) <hazelcast.proxy.queue.Queue.poll>`

        :param timeout: (long), maximum time in seconds to wait for addition (optional).
        :return: (object), the head of this queue, or ``None`` if this queue is empty or specified timeout elapses before an
        item is added to the queue.
        """
        return self._encode_invoke(transactional_queue_poll_codec, timeout=to_millis(timeout))

    def peek(self, timeout=0):
        """
        Transactional implementation of :func:`Queue.peek(timeout) <hazelcast.proxy.queue.Queue.peek>`

        :param timeout: (long), maximum time in seconds to wait for addition (optional).
        :return: (object), the head of this queue, or ``None`` if this queue is empty or specified timeout elapses before an
        item is added to the queue.
        """
        return self._encode_invoke(transactional_queue_peek_codec, timeout=to_millis(timeout))

    def size(self):
        """
        Transactional implementation of :func:`Queue.size() <hazelcast.proxy.queue.Queue.size>`

        :return: (int), size of the queue.
        """
        return self._encode_invoke(transactional_queue_size_codec)



import itertools
import threading
import time
import logging
import hazelcast
from collections import Sequence, Iterable

from hazelcast import six
from hazelcast.six.moves import range
from hazelcast.version import GIT_COMMIT_ID, GIT_COMMIT_DATE, CLIENT_VERSION

DEFAULT_ADDRESS = "127.0.0.1"
DEFAULT_PORT = 5701

MILLISECONDS_IN_SECONDS = 1000
NANOSECONDS_IN_SECONDS = 1e9


def check_not_none(val, message):
    """
    Tests if an argument is not ``None``.

    :param val: (object), the argument tested to see if it is not ``None``.
    :param message: (str), the error message.
    """
    if val is None:
        raise AssertionError(message)


def check_true(val, message):
    """
    Tests whether the provided expression is ``true``.

    :param val: (bool), the expression tested to see if it is ``true``.
    :param message: (str), the error message.
    """
    if not val:
        raise AssertionError(message)


def check_not_negative(val, message):
    """
    Tests if a value is not negative.

    :param val: (Number), the value tested to see if it is not negative.
    :param message: (str), the error message.
    """
    if val < 0:
        raise AssertionError(message)


def check_not_empty(collection, message):
    """
    Tests if a collection is not empty.

    :param collection: (Collection), the collection tested to see if it is not empty.
    :param message: (str), the error message.
    """
    if not collection:
        raise AssertionError(message)


def current_time():
    """
    Returns the current time of the system.

    :return: (float), current time of the system.
    """
    return time.time()


def current_time_in_millis():
    """
    Returns the current time of the system in millis.
    :return: (int), current time of the system in millis.
    """
    return to_millis(current_time())


def thread_id():
    """
    Returns the current thread's id.

    :return: (int), current thread's id.
    """
    return threading.currentThread().ident


def to_millis(seconds):
    """
    Converts the time parameter in seconds to milliseconds.

    :param seconds: (Number), the given time in seconds.
    :return: (int), result of the conversation in milliseconds.
    """
    return int(seconds * MILLISECONDS_IN_SECONDS)


def to_nanos(seconds):
    """
    Converts the time parameter in seconds to nanoseconds.

    :param seconds: (Number), the given time in seconds.
    :return: (int), result of the conversation in nanoseconds.
    """
    return int(seconds * NANOSECONDS_IN_SECONDS)


def validate_type(_type):
    """
    Validates the type.

    :param _type: (Type), the type to be validated.
    """
    if not isinstance(_type, type):
        raise ValueError("Serializer should be an instance of {}".format(_type.__name__))


def validate_serializer(serializer, _type):
    """
    Validates the serializer for given type.

    :param serializer: (Serializer), the serializer to be validated.
    :param _type: (Type), type to be used for serializer validation.
    """
    if not issubclass(serializer, _type):
        raise ValueError("Serializer should be an instance of {}".format(_type.__name__))


class AtomicInteger(object):
    """
    AtomicInteger is an Integer which can work atomically.
    """
    def __init__(self, initial=0):
        self.count = itertools.count(start=initial)

    def get_and_increment(self):
        """
        Returns the current value and increment it.

        :return: (int), current value of AtomicInteger.
        """
        return next(self.count)

    def set(self, value):
        """
        Sets the value of this AtomicInteger.
        :param value: (int), the new value of AtomicInteger.
        """
        self.count = itertools.count(start=value)


def enum(**enums):
    """
    Utility method for defining enums.
    :param enums: Parameters of enumeration.
    :return: (Enum), the created enumerations.
    """
    enums['reverse'] = dict((value, key) for key, value in six.iteritems(enums))
    return type('Enum', (), enums)


def _parse_address(address):
    if ":" in address:
        host, port = address.split(":")
        return [hazelcast.core.Address(host, int(port))]
    return [hazelcast.core.Address(address, p) for p in range(DEFAULT_PORT, DEFAULT_PORT + 3)]


def get_possible_addresses(addresses=[], member_list=[]):
    return set((addresses + [m.address for m in member_list])) or _parse_address(DEFAULT_ADDRESS)


def get_provider_addresses(providers=[]):
    return list(itertools.chain(*[p.load_addresses() for p in providers]))


def parse_addresses(addresses=[]):
    return list(itertools.chain(*[_parse_address(a) for a in addresses]))


class ImmutableLazyDataList(Sequence):
    def __init__(self, list_data, to_object):
        super(ImmutableLazyDataList, self).__init__()
        self._list_data = list_data
        self._list_obj = [None] * len(self._list_data)
        self.to_object = to_object

    def __contains__(self, value):
        return super(ImmutableLazyDataList, self).__contains__(value)

    def __len__(self):
        return self._list_data.__len__()

    def __getitem__(self, index):
        val = self._list_obj[index]
        if not val:
            data = self._list_data[index]
            if isinstance(data, tuple):
                (key, value) = data
                self._list_obj[index] = (self.to_object(key), self.to_object(value))
            else:
                self._list_obj[index] = self.to_object(data)
        return self._list_obj[index]

    def __eq__(self, other):
        if not isinstance(other, Iterable):
            return False
        self._populate()
        return self._list_obj == other

    def _populate(self):
        for index, data in enumerate(self._list_data):
            if not self._list_obj[index]:
                self.__getitem__(index)

    def __repr__(self):
        self._populate()
        return str(self._list_obj)


# Serialization Utilities

def get_portable_version(portable, default_version):
    try:
        version = portable.get_class_version()
    except AttributeError:
        version = default_version
    return version


class TimeUnit(object):
    """
    Represents the time durations at given units in seconds.
    """
    NANOSECOND = 1e-9
    MICROSECOND = 1e-6
    MILLISECOND = 1e-3
    SECOND = 1.0
    MINUTE = 60.0
    HOUR = 3600.0

    @staticmethod
    def to_seconds(value, time_unit):
        """
        :param value: (Number), value to be translated to seconds
        :param time_unit: Time duration in seconds
        :return: Value of the value in seconds
        """
        if isinstance(value, bool):
            # bool is a subclass of int. Don't let bool and float multiplication.
            raise TypeError
        return float(value) * time_unit


# Version utilities
UNKNOWN_VERSION = -1
MAJOR_VERSION_MULTIPLIER = 10000
MINOR_VERSION_MULTIPLIER = 100


def calculate_version(version_str):
    if not version_str:
        return UNKNOWN_VERSION

    main_parts = version_str.split("-")
    tokens = main_parts[0].split(".")

    if len(tokens) < 2:
        return UNKNOWN_VERSION

    try:
        major_coeff = int(tokens[0])
        minor_coeff = int(tokens[1])

        calculated_version = major_coeff * MAJOR_VERSION_MULTIPLIER + minor_coeff * MINOR_VERSION_MULTIPLIER

        if len(tokens) > 2:
            patch_coeff = int(tokens[2])
            calculated_version += patch_coeff

        return calculated_version
    except ValueError:
        return UNKNOWN_VERSION

# Logging utilities


DEFAULT_LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "filters": {
        "version_message_filter": {
            "()": "hazelcast.util.VersionMessageFilter"
        }
    },
    "formatters": {
        "hazelcast_formatter": {
            "()": "hazelcast.util.HazelcastFormatter",
            "format": "%(asctime)s %(name)s\n%(levelname)s: %(version_message)s %(message)s",
            "datefmt": "%b %d, %Y %I:%M:%S %p"
        }
    },
    "handlers": {
        "console_handler": {
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
            "filters": ["version_message_filter"],
            "formatter": "hazelcast_formatter"
        }
    },
    "loggers": {
        "HazelcastClient": {
            "handlers": ["console_handler"]
        }
    }
}


class VersionMessageFilter(logging.Filter):
    def filter(self, record):
        record.version_message = "[" + CLIENT_VERSION + "]"
        return True


class HazelcastFormatter(logging.Formatter):
    def format(self, record):
        client_name = getattr(record, "client_name", None)
        group_name = getattr(record, "group_name", None)
        if client_name and group_name:
            record.msg = "[" + group_name + "] [" + client_name + "] " + record.msg
        return super(HazelcastFormatter, self).format(record)


def create_git_info():
    if GIT_COMMIT_DATE:
        if GIT_COMMIT_ID:
            return "(" + GIT_COMMIT_DATE + " - " + GIT_COMMIT_ID + ") "
        return "(" + GIT_COMMIT_DATE + ") "

    if GIT_COMMIT_ID:
        return "(" + GIT_COMMIT_ID + ") "
    return ""



from hazelcast.protocol.error_codes import *


def retryable(cls):
    """
    Makes the given error retryable.

    :param cls: (:class:`~hazelcast.exception.HazelcastError`), the given error.
    :return: (:class:`~hazelcast.exception.HazelcastError`), the given error with retryable property.
    """
    cls.retryable = True
    return cls


class HazelcastError(Exception):
    """
    General HazelcastError class.
    """
    pass


class ArrayIndexOutOfBoundsError(HazelcastError):
    pass


class ArrayStoreError(HazelcastError):
    pass


class AuthenticationError(HazelcastError):
    pass


class CacheNotExistsError(HazelcastError):
    pass


@retryable
class CallerNotMemberError(HazelcastError):
    pass


class CancellationError(HazelcastError):
    pass


class ClassCastError(HazelcastError):
    pass


class ClassNotFoundError(HazelcastError):
    pass


class ConcurrentModificationError(HazelcastError):
    pass


class ConfigMismatchError(HazelcastError):
    pass


class ConfigurationError(HazelcastError):
    pass


class DistributedObjectDestroyedError(HazelcastError):
    pass


class DuplicateInstanceNameError(HazelcastError):
    pass


class HazelcastEOFError(HazelcastError):
    pass


class ExecutionError(HazelcastError):
    pass


@retryable
class HazelcastInstanceNotActiveError(HazelcastError):
    pass


class HazelcastOverloadError(HazelcastError):
    pass


class HazelcastSerializationError(HazelcastError):
    pass


class HazelcastIOError(HazelcastError):
    pass


class IllegalArgumentError(HazelcastError):
    pass


class IllegalAccessException(HazelcastError):
    pass


class IllegalAccessError(HazelcastError):
    pass


class IllegalMonitorStateError(HazelcastError):
    pass


class HazelcastIllegalStateError(HazelcastError):
    pass


class IllegalThreadStateError(HazelcastError):
    pass


class IndexOutOfBoundsError(HazelcastError):
    pass


class HazelcastInterruptedError(HazelcastError):
    pass


class InvalidAddressError(HazelcastError):
    pass


class InvalidConfigurationError(HazelcastError):
    pass


@retryable
class MemberLeftError(HazelcastError):
    pass


class NegativeArraySizeError(HazelcastError):
    pass


class NoSuchElementError(HazelcastError):
    pass


class NotSerializableError(HazelcastError):
    pass


class NullPointerError(HazelcastError):
    pass


class OperationTimeoutError(HazelcastError):
    pass


@retryable
class PartitionMigratingError(HazelcastError):
    pass


class QueryError(HazelcastError):
    pass


class QueryResultSizeExceededError(HazelcastError):
    pass


class QuorumError(HazelcastError):
    pass


class ReachedMaxSizeError(HazelcastError):
    pass


class RejectedExecutionError(HazelcastError):
    pass


class RemoteMapReduceError(HazelcastError):
    pass


class ResponseAlreadySentError(HazelcastError):
    pass


@retryable
class RetryableHazelcastError(HazelcastError):
    pass


@retryable
class RetryableIOError(HazelcastError):
    pass


class HazelcastRuntimeError(HazelcastError):
    pass


class SecurityError(HazelcastError):
    pass


class SocketError(HazelcastError):
    pass


class StaleSequenceError(HazelcastError):
    pass


class TargetDisconnectedError(HazelcastError):
    pass


@retryable
class TargetNotMemberError(HazelcastError):
    pass


class TimeoutError(HazelcastError):
    pass


class TopicOverloadError(HazelcastError):
    pass


class TopologyChangedError(HazelcastError):
    pass


class TransactionError(HazelcastError):
    pass


class TransactionNotActiveError(HazelcastError):
    pass


class TransactionTimedOutError(HazelcastError):
    pass


class URISyntaxError(HazelcastError):
    pass


class UTFDataFormatError(HazelcastError):
    pass


class UnsupportedOperationError(HazelcastError):
    pass


@retryable
class WrongTargetError(HazelcastError):
    pass


class XAError(HazelcastError):
    pass


class AccessControlError(HazelcastError):
    pass


class LoginError(HazelcastError):
    pass


class UnsupportedCallbackError(HazelcastError):
    pass


class NoDataMemberInClusterError(HazelcastError):
    pass


class ReplicatedMapCantBeCreatedOnLiteMemberError(HazelcastError):
    pass


class MaxMessageSizeExceededError(HazelcastError):
    pass


class WANReplicationQueueFullError(HazelcastError):
    pass


class HazelcastAssertionError(HazelcastError):
    pass


class OutOfMemoryError(HazelcastError):
    pass


class StackOverflowError(HazelcastError):
    pass


class NativeOutOfMemoryError(HazelcastError):
    pass


class ServiceNotFoundError(HazelcastError):
    pass


class StaleTaskIdError(HazelcastError):
    pass


class DuplicateTaskError(HazelcastError):
    pass


class StaleTaskError(HazelcastError):
    pass


class LocalMemberResetError(HazelcastError):
    pass


class IndeterminateOperationStateError(HazelcastError):
    pass


class NodeIdOutOfRangeError(HazelcastError):
    pass


@retryable
class TargetNotReplicaError(HazelcastError):
    pass


class MutationDisallowedError(HazelcastError):
    pass


class ConsistencyLostError(HazelcastError):
    pass


class HazelcastClientNotActiveException(ValueError):
    pass


class HazelcastCertificationError(HazelcastError):
    pass


ERROR_CODE_TO_ERROR = {
    ARRAY_INDEX_OUT_OF_BOUNDS: ArrayIndexOutOfBoundsError,
    ARRAY_STORE: ArrayStoreError,
    AUTHENTICATION: AuthenticationError,
    CACHE_NOT_EXISTS: CacheNotExistsError,
    CALLER_NOT_MEMBER: CallerNotMemberError,
    CANCELLATION: CancellationError,
    CLASS_CAST: ClassCastError,
    CLASS_NOT_FOUND: ClassNotFoundError,
    CONCURRENT_MODIFICATION: ConcurrentModificationError,
    CONFIG_MISMATCH: ConfigMismatchError,
    CONFIGURATION: ConfigurationError,
    DISTRIBUTED_OBJECT_DESTROYED: DistributedObjectDestroyedError,
    DUPLICATE_INSTANCE_NAME: DuplicateInstanceNameError,
    EOF: HazelcastEOFError,
    EXECUTION: ExecutionError,
    HAZELCAST: HazelcastError,
    HAZELCAST_INSTANCE_NOT_ACTIVE: HazelcastInstanceNotActiveError,
    HAZELCAST_OVERLOAD: HazelcastOverloadError,
    HAZELCAST_SERIALIZATION: HazelcastSerializationError,
    IO: HazelcastIOError,
    ILLEGAL_ARGUMENT: IllegalArgumentError,
    ILLEGAL_ACCESS_EXCEPTION: IllegalAccessException,
    ILLEGAL_ACCESS_ERROR: IllegalAccessError,
    ILLEGAL_MONITOR_STATE: IllegalMonitorStateError,
    ILLEGAL_STATE: HazelcastIllegalStateError,
    ILLEGAL_THREAD_STATE: IllegalThreadStateError,
    INDEX_OUT_OF_BOUNDS: IndexOutOfBoundsError,
    INTERRUPTED: HazelcastInterruptedError,
    INVALID_ADDRESS: InvalidAddressError,
    INVALID_CONFIGURATION: InvalidConfigurationError,
    MEMBER_LEFT: MemberLeftError,
    NEGATIVE_ARRAY_SIZE: NegativeArraySizeError,
    NO_SUCH_ELEMENT: NoSuchElementError,
    NOT_SERIALIZABLE: NotSerializableError,
    NULL_POINTER: NullPointerError,
    OPERATION_TIMEOUT: OperationTimeoutError,
    PARTITION_MIGRATING: PartitionMigratingError,
    QUERY: QueryError,
    QUERY_RESULT_SIZE_EXCEEDED: QueryResultSizeExceededError,
    QUORUM: QuorumError,
    REACHED_MAX_SIZE: ReachedMaxSizeError,
    REJECTED_EXECUTION: RejectedExecutionError,
    REMOTE_MAP_REDUCE: RemoteMapReduceError,
    RESPONSE_ALREADY_SENT: ResponseAlreadySentError,
    RETRYABLE_HAZELCAST: RetryableHazelcastError,
    RETRYABLE_IO: RetryableIOError,
    RUNTIME: HazelcastRuntimeError,
    SECURITY: SecurityError,
    SOCKET: SocketError,
    STALE_SEQUENCE: StaleSequenceError,
    TARGET_DISCONNECTED: TargetDisconnectedError,
    TARGET_NOT_MEMBER: TargetNotMemberError,
    TIMEOUT: TimeoutError,
    TOPIC_OVERLOAD: TopicOverloadError,
    TOPOLOGY_CHANGED: TopologyChangedError,
    TRANSACTION: TransactionError,
    TRANSACTION_NOT_ACTIVE: TransactionNotActiveError,
    TRANSACTION_TIMED_OUT: TransactionTimedOutError,
    URI_SYNTAX: URISyntaxError,
    UTF_DATA_FORMAT: UTFDataFormatError,
    UNSUPPORTED_OPERATION: UnsupportedOperationError,
    WRONG_TARGET: WrongTargetError,
    XA: XAError,
    ACCESS_CONTROL: AccessControlError,
    LOGIN: LoginError,
    UNSUPPORTED_CALLBACK: UnsupportedCallbackError,
    NO_DATA_MEMBER: NoDataMemberInClusterError,
    REPLICATED_MAP_CANT_BE_CREATED: ReplicatedMapCantBeCreatedOnLiteMemberError,
    MAX_MESSAGE_SIZE_EXCEEDED: MaxMessageSizeExceededError,
    WAN_REPLICATION_QUEUE_FULL: WANReplicationQueueFullError,
    ASSERTION_ERROR: HazelcastAssertionError,
    OUT_OF_MEMORY_ERROR: OutOfMemoryError,
    STACK_OVERFLOW_ERROR: StackOverflowError,
    NATIVE_OUT_OF_MEMORY_ERROR: NativeOutOfMemoryError,
    SERVICE_NOT_FOUND: ServiceNotFoundError,
    STALE_TASK_ID: StaleTaskIdError,
    DUPLICATE_TASK: DuplicateTaskError,
    STALE_TASK: StaleTaskError,
    LOCAL_MEMBER_RESET: LocalMemberResetError,
    INDETERMINATE_OPERATION_STATE: IndeterminateOperationStateError,
    FLAKE_ID_NODE_ID_OUT_OF_RANGE_EXCEPTION: NodeIdOutOfRangeError,
    TARGET_NOT_REPLICA_EXCEPTION: TargetNotReplicaError,
    MUTATION_DISALLOWED_EXCEPTION: MutationDisallowedError,
    CONSISTENCY_LOST_EXCEPTION: ConsistencyLostError,
}


def create_exception(error_codec):
    """
    Creates an exception with given error codec.

    :param error_codec: (Error Codec), error codec which includes the class name, message and exception trace.
    :return: (Exception), the created exception.
    """
    if error_codec.error_code in ERROR_CODE_TO_ERROR:
        return ERROR_CODE_TO_ERROR[error_codec.error_code](error_codec.message)

    stack_trace = "\n".join(
        ["\tat %s.%s(%s:%s)" % (x.declaring_class, x.method_name, x.file_name, x.line_number) for x in
         error_codec.stack_trace])
    message = "Got exception from server:\n %s: %s\n %s" % (error_codec.class_name,
                                                            error_codec.message,
                                                            stack_trace)
    return HazelcastError(message)


def is_retryable_error(error):
    """
    Determines whether the given error is retryable or not.
    :param error: (:class:`~hazelcast.exception.HazelcastError`), the given error.
    :return: (bool), ``true`` if the given error is retryable, ``false`` otherwise.
    """
    return hasattr(error, 'retryable')



"""Hazelcast Core objects"""
import json

from hazelcast import six
from hazelcast import util


class Member(object):
    """
    Represents a member in the cluster with its address, uuid, lite member status and attributes.
    """
    def __init__(self, address, uuid, is_lite_member=False, attributes={}):
        self.address = address
        self.uuid = uuid
        self.is_lite_member = is_lite_member
        self.attributes = attributes

    def __str__(self):
        return "Member [{}]:{} - {}".format(self.address.host, self.address.port, self.uuid)

    def __repr__(self):
        return "Member(host={}, port={}, uuid={}, liteMember={}, attributes={})" \
            .format(self.address.host, self.address.port, self.uuid, self.is_lite_member, self.attributes)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.address == other.address and self.uuid == other.uuid


class Address(object):
    """
    Represents an address of a member in the cluster.
    """
    def __init__(self, host, port):
        self.host = host
        self.port = port

    def __repr__(self):
        return "Address(host=%s, port=%d)" % (self.host, self.port)

    def __hash__(self):
        return hash((self.host, self.port))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and (self.host, self.port) == (other.host, other.port)


class DistributedObjectInfo(object):
    """
    Represents name of the Distributed Object and the name of service which it belongs to.
    """
    def __init__(self, name, service_name):
        self.name = name
        self.service_name = service_name

    def __repr__(self):
        return "DistributedObjectInfo(name={}, serviceName={})".format(self.name, self.service_name)


class EntryView(object):
    """
    EntryView represents a readonly view of a map entry.
    """
    key = None
    """
    The key of the entry.
    """
    value = None
    """
    The value of the entry.
    """
    cost = None
    """
    The cost in bytes of the entry.
    """
    creation_time = None
    """
    The creation time of the entry.
    """
    expiration_time = None
    """
    The expiration time of the entry.
    """
    hits = None
    """
    Number of hits of the entry.
    """
    last_access_time = None
    """
    The last access time for the entry.
    """
    last_stored_time = None
    """
    The last store time for the value.
    """
    last_update_time = None
    """
    The last time the value was updated.
    """
    version = None
    """
    The version of the entry.
    """
    eviction_criteria_number = None
    """
    The criteria number for eviction.
    """
    ttl = None
    """
    The last set time to live second.
    """

    def __repr__(self):
        return "EntryView(key=%s, value=%s, cost=%s, creation_time=%s, expiration_time=%s, hits=%s, last_access_time=%s, " \
               "last_stored_time=%s, last_update_time=%s, version=%s, eviction_criteria_number=%s, ttl=%s" % (
                   self.key, self.value, self.cost, self.creation_time, self.expiration_time, self.hits,
                   self.last_access_time, self.last_stored_time, self.last_update_time, self.version,
                   self.eviction_criteria_number, self.ttl)


class MemberSelector(object):
    """
    Subclasses of this class select members
    that are capable of executing a special kind of task.
    The select(Member) method is called for every available
    member in the cluster and it is up to the implementation to decide
    if the member is going to be used or not.
    """
    def select(self, member):
        """
        Decides if the given member will be part of an operation or not.

        :param member: (:class:`~hazelcast.core.Member`), the member instance to decide upon.
        :return: (bool), True if the member should take part in the operation, False otherwise.
        """
        raise NotImplementedError()


class DataMemberSelector(MemberSelector):
    def select(self, member):
        return not member.is_lite_member


class HazelcastJsonValue(object):
    """
    HazelcastJsonValue is a wrapper for JSON formatted strings. It is preferred
    to store HazelcastJsonValue instead of Strings for JSON formatted strings.
    Users can run predicates and use indexes on the attributes of the underlying
    JSON strings.

    HazelcastJsonValue is queried using Hazelcast's querying language.
    See `Distributed Query section <https://github.com/hazelcast/hazelcast-python-client#77-distributed-query>`_.

    In terms of querying, numbers in JSON strings are treated as either
    Long or Double in the Java side. str, bool and None
    are treated as String, boolean and null respectively.

    HazelcastJsonValue keeps given string as it is. Strings are not
    checked for being valid. Ill-formatted JSON strings may cause false
    positive or false negative results in queries.

    HazelcastJsonValue can also be constructed from JSON serializable objects.
    In that case, objects are converted to JSON strings and stored as such.
    If an error occurs during the conversion, it is raised directly.

    None values are not allowed.
    """
    def __init__(self, value):
        util.check_not_none(value, "JSON string or the object cannot be None.")
        if isinstance(value, six.string_types):
            self._json_string = value
        else:
            self._json_string = json.dumps(value)

    def to_string(self):
        """
        Returns unaltered string that was used to create this object.

        :return: (str), original string
        """
        return self._json_string

    def loads(self):
        """
        Deserializes the string that was used to create this object
        and returns as Python object.

        :return: (object), Python object represented by the original string
        """
        return json.loads(self._json_string)




CLIENT_AUTHENTICATION = 0x0002
CLIENT_AUTHENTICATIONCUSTOM = 0x0003
CLIENT_ADDMEMBERSHIPLISTENER = 0x0004
CLIENT_CREATEPROXY = 0x0005
CLIENT_DESTROYPROXY = 0x0006
CLIENT_GETPARTITIONS = 0x0008
CLIENT_REMOVEALLLISTENERS = 0x0009
CLIENT_ADDPARTITIONLOSTLISTENER = 0x000a
CLIENT_REMOVEPARTITIONLOSTLISTENER = 0x000b
CLIENT_GETDISTRIBUTEDOBJECTS = 0x000c
CLIENT_ADDDISTRIBUTEDOBJECTLISTENER = 0x000d
CLIENT_REMOVEDISTRIBUTEDOBJECTLISTENER = 0x000e
CLIENT_PING = 0x000f
CLIENT_STATISTICS = 0x0010



"""
Constants
"""
BYTE_SIZE_IN_BYTES = 1
BOOLEAN_SIZE_IN_BYTES = 1
SHORT_SIZE_IN_BYTES = 2
CHAR_SIZE_IN_BYTES = 2
INT_SIZE_IN_BYTES = 4
FLOAT_SIZE_IN_BYTES = 4
LONG_SIZE_IN_BYTES = 8
DOUBLE_SIZE_IN_BYTES = 8

FMT_LE_INT = "<i"
FMT_LE_INT8 = "<b"
FMT_LE_UINT8 = "<B"
FMT_LE_INT16 = "<h"
FMT_LE_UINT16 = "<H"
FMT_LE_LONG = "<q"
FMT_LE_FLOAT = "<f"
FMT_LE_DOUBLE = "<d"

FMT_BE_INT = ">i"
FMT_BE_INT8 = ">b"
FMT_BE_UINT8 = ">B"
FMT_BE_INT16 = ">h"
FMT_BE_UINT16 = ">H"
FMT_BE_LONG = ">q"
FMT_BE_FLOAT = ">f"
FMT_BE_DOUBLE = ">d"

BIG_ENDIAN = 2
LITTLE_ENDIAN = 1

NULL_ARRAY_LENGTH = -1

# LIMITS
MAX_BYTE = 127
MIN_BYTE = -128

MAX_SHORT = 2 ** 16 - 1
MIN_SHORT = -2 ** 15

MAX_INT = 2 ** 32 - 1
MIN_INT = -2 ** 31

MAX_LONG = 2 ** 64 - 1
MIN_LONG = -2 ** 63

MAX_FLOAT32 = 3.4028235e+38
MIN_FLOAT32 = 1.4e-45


def calculate_size_str(val):
    return len(val) + INT_SIZE_IN_BYTES


def calculate_size_data(val):
    return len(val) + INT_SIZE_IN_BYTES


def calculate_size_address(val):
    return calculate_size_str(val.host) + INT_SIZE_IN_BYTES



import threading

from hazelcast.proxy.base import Proxy
from hazelcast.util import AtomicInteger

BLOCK_SIZE = 10000


class IdGenerator(Proxy):
    """
    The IdGenerator is responsible for creating unique ids (a long) in a cluster. In theory, an
    AtomicLong.increment_and_get() could be used to provide the same functionality. The big difference is that the
    increment_and_get() requires one or more remote calls for every invocation which cause a performance and
    scalability bottleneck. The IdGenerator uses an AtomicLong under the hood, but instead of doing remote call for
    every call to new_id(), it does it less frequently. It checks out a chunk, e.g. 1..1000 and as long as it has not
    yet consumed all the ids in its chunk, then no remote call is done. IDs generated by different cluster members may
    get out of order because each member will get its own chunk. It can be that member 1 has chunk 1..1000 and member
    2 has 1001..2000. Therefore, member 2 will automatically have ids that are out of order with the ids generated by
    member 1.
    """
    def __init__(self, client, service_name, name, atomic_long):
        super(IdGenerator, self).__init__(client, service_name, name)
        self._atomic_long = atomic_long
        self._residue = BLOCK_SIZE
        self._local = -1
        self._lock = threading.RLock()

    def _on_destroy(self):
        self._atomic_long.destroy()

    def init(self, initial):
        """
        Try to initialize this IdGenerator instance with the given id. The first generated id will be 1 greater than id.

        :param initial: (long), the given id.
        :return: (bool), ``true`` if initialization succeeded, ``false`` if id is less than 0.
        """
        if initial <= 0:
            return False
        step = initial // BLOCK_SIZE
        with self._lock:
            init = self._atomic_long.compare_and_set(0, step + 1).result()
            if init:
                self._local = step
                self._residue = (initial % BLOCK_SIZE) + 1
            return init

    def new_id(self):
        """
        Generates and returns a cluster-wide unique id. Generated ids are guaranteed to be unique for the entire cluster
        as long as the cluster is live. If the cluster restarts, then id generation will start from 0.

        :return: (long), cluster-wide new unique id.
        """
        with self._lock:
            curr = self._residue
            self._residue += 1
            if self._residue >= BLOCK_SIZE:
                increment = self._atomic_long.get_and_increment().result()
                self._local = increment
                self._residue = 0
                return self.new_id()
            return self._local * BLOCK_SIZE + curr




from hazelcast.client import HazelcastClient
from hazelcast.config import ClientConfig, ClientNetworkConfig, SerializationConfig, GroupConfig, SSLConfig, \
    ClientCloudConfig, FlakeIdGeneratorConfig
from hazelcast.version import CLIENT_VERSION_INFO as __version_info__
from hazelcast.version import CLIENT_VERSION as __version__




EXECUTORSERVICE_SHUTDOWN = 0x0901
EXECUTORSERVICE_ISSHUTDOWN = 0x0902
EXECUTORSERVICE_CANCELONPARTITION = 0x0903
EXECUTORSERVICE_CANCELONADDRESS = 0x0904
EXECUTORSERVICE_SUBMITTOPARTITION = 0x0905
EXECUTORSERVICE_SUBMITTOADDRESS = 0x0906



import struct

from hazelcast.serialization.api import *
from hazelcast.serialization.bits import *
from hazelcast.six.moves import range


class _ObjectDataOutput(ObjectDataOutput):
    def __init__(self, init_size, serialization_service, is_big_endian=True):
        self._init_size = init_size
        self._buffer = bytearray(init_size)
        self._service = serialization_service
        self._is_big_endian = is_big_endian
        self._pos = 0
        # Local cache struct formats according to endianness
        self._FMT_INT = FMT_BE_INT if self._is_big_endian else FMT_LE_INT
        self._FMT_SHORT = FMT_BE_INT16 if self._is_big_endian else FMT_LE_INT16
        self._CHAR_ENCODING = "utf_16_be" if self._is_big_endian else "utf_16_le"
        self._FMT_LONG = FMT_BE_LONG if self._is_big_endian else FMT_LE_LONG
        self._FMT_FLOAT = FMT_BE_FLOAT if self._is_big_endian else FMT_LE_FLOAT
        self._FMT_DOUBLE = FMT_BE_DOUBLE if self._is_big_endian else FMT_LE_DOUBLE

    def _write(self, val):
        self._ensure_available(BYTE_SIZE_IN_BYTES)
        self._buffer[self._pos] = val
        self._pos += BYTE_SIZE_IN_BYTES

    def write_from(self, buff, offset=None, length=None):
        _off = offset if offset is not None else 0
        _len = length if length is not None else len(buff)
        if _off < 0 or _len < 0 or (_off + _len) > len(buff):
            raise IndexError()
        elif length == 0:
            return
        self._ensure_available(_len)
        self._buffer[self._pos: self._pos + _len] = buff[:]
        self._pos += _len

    def write_boolean(self, boolean):
        self._write(1 if boolean else 0)

    def write_byte(self, val):
        self._write(val)

    def write_short(self, val):
        self._ensure_available(SHORT_SIZE_IN_BYTES)
        struct.pack_into(self._FMT_SHORT, self._buffer, self._pos, val)
        self._pos += SHORT_SIZE_IN_BYTES

    def write_char(self, val):
        encoded = val.encode(self._CHAR_ENCODING)
        self.write_from(encoded)

    def write_int(self, val, position=None):
        self._ensure_available(INT_SIZE_IN_BYTES)
        if position is None:
            struct.pack_into(self._FMT_INT, self._buffer, self._pos, val)
            self._pos += INT_SIZE_IN_BYTES
        else:
            struct.pack_into(self._FMT_INT, self._buffer, position, val)

    def write_int_big_endian(self, val):
        self._ensure_available(INT_SIZE_IN_BYTES)
        struct.pack_into(FMT_BE_INT, self._buffer, self._pos, val)
        self._pos += INT_SIZE_IN_BYTES

    def write_long(self, val):
        self._ensure_available(LONG_SIZE_IN_BYTES)
        struct.pack_into(self._FMT_LONG, self._buffer, self._pos, val)
        self._pos += LONG_SIZE_IN_BYTES

    def write_float(self, val):
        self._ensure_available(FLOAT_SIZE_IN_BYTES)
        struct.pack_into(self._FMT_FLOAT, self._buffer, self._pos, val)
        self._pos += FLOAT_SIZE_IN_BYTES

    def write_double(self, val):
        self._ensure_available(DOUBLE_SIZE_IN_BYTES)
        struct.pack_into(self._FMT_DOUBLE, self._buffer, self._pos, val)
        self._pos += DOUBLE_SIZE_IN_BYTES

    def write_utf(self, val):
        _len = len(val) if val is not None else NULL_ARRAY_LENGTH
        self.write_int(_len)
        if _len > 0:
            self.write_from(val.encode("utf-8"))

    def write_byte_array(self, val):
        _len = len(val) if val is not None else NULL_ARRAY_LENGTH
        self.write_int(_len)
        if _len > 0:
            self.write_from(val)

    def write_boolean_array(self, val):
        self._write_array_fnc(val, self.write_boolean)

    def write_char_array(self, val):
        self._write_array_fnc(val, self.write_char)

    def write_int_array(self, val):
        self._write_array_fnc(val, self.write_int)

    def write_long_array(self, val):
        self._write_array_fnc(val, self.write_long)

    def write_double_array(self, val):
        self._write_array_fnc(val, self.write_double)

    def write_float_array(self, val):
        self._write_array_fnc(val, self.write_float)

    def write_short_array(self, val):
        self._write_array_fnc(val, self.write_short)

    def write_utf_array(self, val):
        self._write_array_fnc(val, self.write_utf)

    def write_object(self, val):
        self._service.write_object(self, val)

    def write_data(self, data):
        payload = data.to_bytes() if data is not None else None
        self.write_byte_array(payload)

    def to_byte_array(self):
        if self._buffer is None or self._pos == 0:
            return bytearray()
        new_buffer = bytearray(self._pos)
        new_buffer[:] = self._buffer[:self._pos]
        return new_buffer

    def is_big_endian(self):
        return self._is_big_endian

    def position(self):
        return self._pos

    def set_position(self, position):
        self._pos = position

    def write_zero_bytes(self, count):
        for i in range(0, count):
            self._write(0)

    # HELPERS
    def _write_array_fnc(self, val, item_write_fnc):
        _len = len(val) if val is not None else NULL_ARRAY_LENGTH
        self.write_int(_len)
        if _len > 0:
            for item in val:
                item_write_fnc(item)

    def _ensure_available(self, length):
        if self._available() < length:
            if self._buffer is not None:
                buffer_length = len(self._buffer)
                new_length = max(buffer_length << 1, buffer_length + length)
                new_buffer = bytearray(new_length)
                new_buffer[:self._pos] = self._buffer[:self._pos]
                self._buffer = new_buffer
            else:
                new_length = length * 2 if length > self._init_size // 2 else self._init_size
                self._buffer = bytearray(new_length)

    def _available(self):
        return len(self._buffer) - self._pos if self._buffer is not None else 0

    def __repr__(self):
        from binascii import hexlify
        buf = hexlify(self._buffer)
        pos_ = self._pos * 2
        return buf[:pos_] + "[" + buf[pos_] + "]" + buf[pos_ + 1:]


class EmptyObjectDataOutput(ObjectDataOutput):
    def write_utf_array(self, val):
        pass

    def write_utf(self, val):
        pass

    def write_short_array(self, val):
        pass

    def write_short(self, val):
        pass

    def write_object(self, val):
        pass

    def write_long_array(self, val):
        pass

    def write_long(self, val):
        pass

    def write_int_array(self, val):
        pass

    def write_int(self, val):
        pass

    def write_from(self, buff, offset=None, length=None):
        pass

    def write_float_array(self, val):
        pass

    def write_float(self, val):
        pass

    def write_double_array(self, val):
        pass

    def write_double(self, val):
        pass

    def write_data(self, val):
        pass

    def write_chars(self, val):
        pass

    def write_char_array(self, val):
        pass

    def write_char(self, val):
        pass

    def write_bytes(self, string):
        pass

    def write_byte_array(self, val):
        pass

    def write_byte(self, val):
        pass

    def write_boolean_array(self, val):
        pass

    def write_boolean(self, val):
        pass

    def to_byte_array(self):
        raise NotImplementedError("to_byte_array not implemented")

    def get_byte_order(self):
        pass



from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import ClientMessage
from hazelcast.protocol.custom_codec import *
from hazelcast.util import ImmutableLazyDataList
from hazelcast.protocol.codec.client_message_type import *
from hazelcast.six.moves import range

REQUEST_TYPE = CLIENT_AUTHENTICATION
RESPONSE_TYPE = 107
RETRYABLE = True


def calculate_size(username, password, uuid, owner_uuid, is_owner_connection, client_type, serialization_version, client_hazelcast_version):
    """ Calculates the request payload size"""
    data_size = 0
    data_size += calculate_size_str(username)
    data_size += calculate_size_str(password)
    data_size += BOOLEAN_SIZE_IN_BYTES
    if uuid is not None:
        data_size += calculate_size_str(uuid)
    data_size += BOOLEAN_SIZE_IN_BYTES
    if owner_uuid is not None:
        data_size += calculate_size_str(owner_uuid)
    data_size += BOOLEAN_SIZE_IN_BYTES
    data_size += calculate_size_str(client_type)
    data_size += BYTE_SIZE_IN_BYTES
    data_size += calculate_size_str(client_hazelcast_version)
    return data_size


def encode_request(username, password, uuid, owner_uuid, is_owner_connection, client_type, serialization_version, client_hazelcast_version):
    """ Encode request into client_message"""
    client_message = ClientMessage(payload_size=calculate_size(username, password, uuid, owner_uuid, is_owner_connection, client_type, serialization_version, client_hazelcast_version))
    client_message.set_message_type(REQUEST_TYPE)
    client_message.set_retryable(RETRYABLE)
    client_message.append_str(username)
    client_message.append_str(password)
    client_message.append_bool(uuid is None)
    if uuid is not None:
        client_message.append_str(uuid)
    client_message.append_bool(owner_uuid is None)
    if owner_uuid is not None:
        client_message.append_str(owner_uuid)
    client_message.append_bool(is_owner_connection)
    client_message.append_str(client_type)
    client_message.append_byte(serialization_version)
    client_message.append_str(client_hazelcast_version)
    client_message.update_frame_length()
    return client_message


def decode_response(client_message, to_object=None):
    """ Decode response from client message"""
    parameters = dict(status=None, address=None, uuid=None, owner_uuid=None, serialization_version=None, server_hazelcast_version=None, client_unregistered_members=None)
    parameters['status'] = client_message.read_byte()
    if not client_message.read_bool():
        parameters['address'] = AddressCodec.decode(client_message, to_object)
    if not client_message.read_bool():
        parameters['uuid'] = client_message.read_str()
    if not client_message.read_bool():
        parameters['owner_uuid'] = client_message.read_str()
    parameters['serialization_version'] = client_message.read_byte()
    if client_message.is_complete():
        return parameters
    parameters['server_hazelcast_version'] = client_message.read_str()
    if not client_message.read_bool():
        client_unregistered_members_size = client_message.read_int()
        client_unregistered_members = []
        for _ in range(0, client_unregistered_members_size):
            client_unregistered_members_item = MemberCodec.decode(client_message, to_object)
            client_unregistered_members.append(client_unregistered_members_item)
        parameters['client_unregistered_members'] = ImmutableLazyDataList(client_unregistered_members, to_object)
    return parameters



from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import ClientMessage
from hazelcast.protocol.custom_codec import *
from hazelcast.util import ImmutableLazyDataList
from hazelcast.protocol.codec.pn_counter_message_type import *
from hazelcast.six.moves import range

REQUEST_TYPE = PNCOUNTER_GET
RESPONSE_TYPE = 127
RETRYABLE = True


def calculate_size(name, replica_timestamps, target_replica):
    """ Calculates the request payload size"""
    data_size = 0
    data_size += calculate_size_str(name)
    data_size += INT_SIZE_IN_BYTES
    for replica_timestamps_item in replica_timestamps:
        key = replica_timestamps_item[0]
        val = replica_timestamps_item[1]
        data_size += calculate_size_str(key)
        data_size += LONG_SIZE_IN_BYTES

    data_size += calculate_size_address(target_replica)
    return data_size


def encode_request(name, replica_timestamps, target_replica):
    """ Encode request into client_message"""
    client_message = ClientMessage(payload_size=calculate_size(name, replica_timestamps, target_replica))
    client_message.set_message_type(REQUEST_TYPE)
    client_message.set_retryable(RETRYABLE)
    client_message.append_str(name)
    client_message.append_int(len(replica_timestamps))
    for replica_timestamps_item in replica_timestamps:
        key = replica_timestamps_item[0]
        val = replica_timestamps_item[1]
        client_message.append_str(key)
        client_message.append_long(val)

    AddressCodec.encode(client_message, target_replica)
    client_message.update_frame_length()
    return client_message


def decode_response(client_message, to_object=None):
    """ Decode response from client message"""
    parameters = dict(value=None, replica_timestamps=None, replica_count=None)
    parameters['value'] = client_message.read_long()

    replica_timestamps_size = client_message.read_int()
    replica_timestamps = []
    for _ in range(0, replica_timestamps_size):
        replica_timestamps_item = (client_message.read_str(), client_message.read_long())
        replica_timestamps.append(replica_timestamps_item)

    parameters['replica_timestamps'] = ImmutableLazyDataList(replica_timestamps, to_object)
    parameters['replica_count'] = client_message.read_int()
    return parameters



from hazelcast.protocol.codec import atomic_reference_compare_and_set_codec, atomic_reference_clear_codec, \
    atomic_reference_contains_codec, atomic_reference_get_and_set_codec, atomic_reference_set_and_get_codec, \
    atomic_reference_get_codec, atomic_reference_is_null_codec, atomic_reference_set_codec, \
    atomic_reference_alter_and_get_codec, atomic_reference_alter_codec, atomic_reference_apply_codec, \
    atomic_reference_get_and_alter_codec
from hazelcast.proxy.base import PartitionSpecificProxy
from hazelcast.util import check_not_none


class AtomicReference(PartitionSpecificProxy):
    """
    AtomicReference is a atomically updated reference to an object.
    """
    def alter(self, function):
        """
        Alters the currently stored reference by applying a function on it.

        :param function: (Function), A stateful serializable object which represents the Function defined on
            server side.
            This object must have a serializable Function counter part registered on server side with the actual
            ``org.hazelcast.core.IFunction`` implementation.
        """
        check_not_none(function, "function can't be None")
        return self._encode_invoke(atomic_reference_alter_codec, function=self._to_data(function))

    def apply(self, function):
        """
        Applies a function on the value, the actual stored value will not change.

        :param function: (Function), A stateful serializable object which represents the Function defined on
            server side.
            This object must have a serializable Function counter part registered on server side with the actual
            ``org.hazelcast.core.IFunction`` implementation.
        :return: (object), the result of the function application.
        """
        check_not_none(function, "function can't be None")
        return self._encode_invoke(atomic_reference_apply_codec, function=self._to_data(function))

    def alter_and_get(self, function):
        """
        Alters the currently stored reference by applying a function on it and gets the result.

        :param function: (Function), A stateful serializable object which represents the Function defined on
            server side.
            This object must have a serializable Function counter part registered on server side with the actual
            ``org.hazelcast.core.IFunction`` implementation.
        :return: (object), the new value, the result of the applied function.
        """
        check_not_none(function, "function can't be None")
        return self._encode_invoke(atomic_reference_alter_and_get_codec, function=self._to_data(function))

    def compare_and_set(self, expected, updated):
        """
        Atomically sets the value to the given updated value only if the current value == the expected value.

        :param expected: (object), the expected value.
        :param updated: (object), the new value.
        :return: (bool), ``true`` if successful; or ``false`` if the actual value was not equal to the expected value.
        """
        return self._encode_invoke(atomic_reference_compare_and_set_codec,
                                   expected=self._to_data(expected), updated=self._to_data(updated))

    def clear(self):
        """
        Clears the current stored reference.
        """
        return self._encode_invoke(atomic_reference_clear_codec)

    def contains(self, expected):
        """
        Checks if the reference contains the value.

        :param expected: (object), the value to check (is allowed to be ``None``).
        :return: (bool), ``true`` if the value is found, ``false`` otherwise.
        """

        return self._encode_invoke(atomic_reference_contains_codec,
                                   expected=self._to_data(expected))

    def get(self):
        """
        Gets the current value.

        :return: (object), the current value.
        """
        return self._encode_invoke(atomic_reference_get_codec)

    def get_and_alter(self, function):
        """
        Alters the currently stored reference by applying a function on it on and gets the old value.

        :param function: (Function), A stateful serializable object which represents the Function defined on
            server side.
            This object must have a serializable Function counter part registered on server side with the actual
            ``org.hazelcast.core.IFunction`` implementation.
        :return: (object), the old value, the value before the function is applied.
        """

        check_not_none(function, "function can't be None")
        return self._encode_invoke(atomic_reference_get_and_alter_codec, function=self._to_data(function))

    def get_and_set(self, new_value):
        """
        Gets the old value and sets the new value.

        :param new_value: (object), the new value.
        :return: (object), the old value.
        """
        return self._encode_invoke(atomic_reference_get_and_set_codec,
                                   new_value=self._to_data(new_value))

    def is_null(self):
        """
        Checks if the stored reference is null.

        :return: (bool), ``true`` if null, ``false`` otherwise.
        """
        return self._encode_invoke(atomic_reference_is_null_codec)

    def set(self, new_value):
        """
        Atomically sets the given value.

        :param new_value: (object), the new value.
        """
        return self._encode_invoke(atomic_reference_set_codec,
                                   new_value=self._to_data(new_value))

    def set_and_get(self, new_value):
        """
        Sets and gets the value.

        :param new_value: (object), the new value.
        :return: (object), the new value.
        """
        return self._encode_invoke(atomic_reference_set_and_get_codec,
                                   new_value=self._to_data(new_value))



"""
Client Message is the carrier framed data as defined below.
Any request parameter, response or event data will be carried in the payload.

0                   1                   2                   3
0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|R|                      Frame Length                           |
+-------------+---------------+---------------------------------+
|  Version    |B|E|  Flags    |               Type              |
+-------------+---------------+---------------------------------+
|                                                               |
+                       CorrelationId                           +
|                                                               |
+---------------------------------------------------------------+
|                        PartitionId                            |
+-----------------------------+---------------------------------+
|        Data Offset          |                                 |
+-----------------------------+                                 |
|                      Message Payload Data                    ...
|                                                              ...


"""
import binascii
import struct
import socket
import errno

from hazelcast.serialization.data import *

# constants
VERSION = 0
BEGIN_FLAG = 0x80
END_FLAG = 0x40
BEGIN_END_FLAG = BEGIN_FLAG | END_FLAG
LISTENER_FLAG = 0x01

PAYLOAD_OFFSET = 18
SIZE_OFFSET = 0

FRAME_LENGTH_FIELD_OFFSET = 0
VERSION_FIELD_OFFSET = FRAME_LENGTH_FIELD_OFFSET + INT_SIZE_IN_BYTES
FLAGS_FIELD_OFFSET = VERSION_FIELD_OFFSET + BYTE_SIZE_IN_BYTES
TYPE_FIELD_OFFSET = FLAGS_FIELD_OFFSET + BYTE_SIZE_IN_BYTES
CORRELATION_ID_FIELD_OFFSET = TYPE_FIELD_OFFSET + SHORT_SIZE_IN_BYTES
PARTITION_ID_FIELD_OFFSET = CORRELATION_ID_FIELD_OFFSET + LONG_SIZE_IN_BYTES
DATA_OFFSET_FIELD_OFFSET = PARTITION_ID_FIELD_OFFSET + INT_SIZE_IN_BYTES
HEADER_SIZE = DATA_OFFSET_FIELD_OFFSET + SHORT_SIZE_IN_BYTES


class ClientMessage(object):
    def __init__(self, buff=None, payload_size=0):
        if buff:
            self.buffer = buff
            self._read_index = 0
        else:
            self.buffer = bytearray(HEADER_SIZE + payload_size)
            self.set_data_offset(HEADER_SIZE)
            self._write_index = 0
        self._retryable = False

    # HEADER ACCESSORS
    def get_correlation_id(self):
        return struct.unpack_from(FMT_LE_LONG, self.buffer, CORRELATION_ID_FIELD_OFFSET)[0]

    def set_correlation_id(self, val):
        struct.pack_into(FMT_LE_LONG, self.buffer, CORRELATION_ID_FIELD_OFFSET, val)
        return self

    def get_partition_id(self):
        return struct.unpack_from(FMT_LE_INT, self.buffer, PARTITION_ID_FIELD_OFFSET)[0]

    def set_partition_id(self, val):
        struct.pack_into(FMT_LE_INT, self.buffer, PARTITION_ID_FIELD_OFFSET, val)
        return self

    def get_message_type(self):
        return struct.unpack_from(FMT_LE_UINT16, self.buffer, TYPE_FIELD_OFFSET)[0]

    def set_message_type(self, val):
        struct.pack_into(FMT_LE_UINT16, self.buffer, TYPE_FIELD_OFFSET, val)
        return self

    def get_flags(self):
        return struct.unpack_from(FMT_LE_UINT8, self.buffer, FLAGS_FIELD_OFFSET)[0]

    def set_flags(self, val):
        struct.pack_into(FMT_LE_UINT8, self.buffer, FLAGS_FIELD_OFFSET, val)
        return self

    def has_flags(self, flags):
        return self.get_flags() & flags

    def get_frame_length(self):
        return struct.unpack_from(FMT_LE_INT, self.buffer, FRAME_LENGTH_FIELD_OFFSET)[0]

    def set_frame_length(self, val):
        struct.pack_into(FMT_LE_INT, self.buffer, FRAME_LENGTH_FIELD_OFFSET, val)
        return self

    def get_data_offset(self):
        return struct.unpack_from(FMT_LE_UINT16, self.buffer, DATA_OFFSET_FIELD_OFFSET)[0]

    def set_data_offset(self, val):
        struct.pack_into(FMT_LE_UINT16, self.buffer, DATA_OFFSET_FIELD_OFFSET, val)
        return self

    def _write_offset(self):
        return self.get_data_offset() + self._write_index

    def _read_offset(self):
        return self.get_data_offset() + self._read_index

    # PAYLOAD
    def append_byte(self, val):
        struct.pack_into(FMT_LE_UINT8, self.buffer, self._write_offset(), val)
        self._write_index += BYTE_SIZE_IN_BYTES
        return self

    def append_bool(self, val):
        return self.append_byte(1 if val else 0)

    def append_int(self, val):
        struct.pack_into(FMT_LE_INT, self.buffer, self._write_offset(), val)
        self._write_index += INT_SIZE_IN_BYTES
        return self

    def append_long(self, val):
        struct.pack_into(FMT_LE_LONG, self.buffer, self._write_offset(), val)
        self._write_index += LONG_SIZE_IN_BYTES
        return self

    def append_str(self, val):
        self.append_byte_array(val.encode("utf-8"))
        return self

    def append_data(self, val):
        self.append_byte_array(val.to_bytes())
        return self

    def append_byte_array(self, arr):
        length = len(arr)
        # length
        self.append_int(length)
        # copy content
        self.buffer[self._write_offset(): self._write_offset() + length] = arr[:]
        self._write_index += length

    def append_tuple(self, entry_tuple):
        self.append_data(entry_tuple[0]).append_data(entry_tuple[1])
        return self

    # PAYLOAD READ
    def _read_from_buff(self, fmt, size):
        val = struct.unpack_from(fmt, self.buffer, self._read_offset())
        self._read_index += size
        return val[0]

    def read_byte(self):
        return self._read_from_buff(FMT_LE_UINT8, BYTE_SIZE_IN_BYTES)

    def read_bool(self):
        return True if self.read_byte() else False

    def read_int(self):
        return self._read_from_buff(FMT_LE_INT, INT_SIZE_IN_BYTES)

    def read_long(self):
        return self._read_from_buff(FMT_LE_LONG, LONG_SIZE_IN_BYTES)

    def read_str(self):
        return self.read_byte_array().decode("utf-8")

    def read_data(self):
        return Data(self.read_byte_array())

    def read_byte_array(self):
        length = self.read_int()
        result = bytearray(self.buffer[self._read_offset(): self._read_offset() + length])
        self._read_index += length
        return result

    # helpers

    def is_retryable(self):
        return self._retryable

    def set_retryable(self, val):
        self._retryable = val
        return self

    def is_complete(self):
        try:
            return (self._read_offset() >= HEADER_SIZE) and (self._read_offset() == self.get_frame_length())
        except AttributeError:
            return False

    def is_flag_set(self, flag):
        i = self.get_flags() & flag
        return i == flag

    def add_flag(self, flags):
        self.set_flags(self.get_flags() | flags)
        return self

    def update_frame_length(self):
        self.set_frame_length(self._write_offset())
        return self

    def accumulate(self, client_message):
        start = client_message.get_data_offset()
        end = client_message.get_frame_length()
        self.buffer += client_message.buffer[start:end]
        self.set_frame_length(len(self.buffer))

    def clone(self):
        client_message = ClientMessage(bytearray(self.buffer))
        client_message.set_retryable(self._retryable)
        return client_message

    def __repr__(self):
        return binascii.hexlify(self.buffer)

    def __str__(self):
        return "ClientMessage:{{" \
               "length={}, " \
               "correlationId={}, " \
               "messageType={}, " \
               "partitionId={}, " \
               "isComplete={}, " \
               "isRetryable={}, " \
               "isEvent={}, " \
               "writeOffset={}}}".format(self.get_frame_length(),
                                         self.get_correlation_id(),
                                         self.get_message_type(),
                                         self.get_partition_id(),
                                         self.is_complete(),
                                         self.is_retryable(),
                                         self.is_flag_set(LISTENER_FLAG),
                                         self.get_data_offset())


class ClientMessageBuilder(object):
    def __init__(self, message_callback):
        self._incomplete_messages = dict()
        self._message_callback = message_callback

    def on_message(self, client_message):
        if client_message.is_flag_set(BEGIN_END_FLAG):
            # handle message
            self._message_callback(client_message)
        elif client_message.is_flag_set(BEGIN_FLAG):
            self._incomplete_messages[client_message.get_correlation_id()] = client_message
        else:
            try:
                message = self._incomplete_messages[client_message.get_correlation_id()]
            except KeyError:
                raise socket.error(errno.EIO, "A message without the begin part is received.")
            message.accumulate(client_message)
            if client_message.is_flag_set(END_FLAG):
                message.add_flag(BEGIN_END_FLAG)
                self._message_callback(message)
                del self._incomplete_messages[client_message.get_correlation_id()]



import logging
import threading

from hazelcast.hash import hash_to_index
from hazelcast.protocol.codec import client_get_partitions_codec
from hazelcast import six

PARTITION_UPDATE_INTERVAL = 10


class PartitionService(object):
    """
    An SPI service for accessing partition related information.
    """
    logger = logging.getLogger("HazelcastClient.PartitionService")
    timer = None

    def __init__(self, client):
        self._client = client
        self._logger_extras = {"client_name": client.name, "group_name": client.config.group_config.name}
        self.partitions = {}

    def start(self):
        """
        Starts the partition service.
        """
        self.logger.debug("Starting partition service", extra=self._logger_extras)

        def partition_updater():
            self._do_refresh()
            self.timer = self._client.reactor.add_timer(PARTITION_UPDATE_INTERVAL, partition_updater)

        self.timer = self._client.reactor.add_timer(PARTITION_UPDATE_INTERVAL, partition_updater)

    def shutdown(self):
        """
        Shutdowns the partition service.
        """
        if self.timer:
            self.timer.cancel()

    def refresh(self):
        """
        Refreshes the partition service.
        """
        self._client.reactor.add_timer(0, self._do_refresh)

    def get_partition_owner(self, partition_id):
        """
        Gets the owner of the partition if it's set. Otherwise it will trigger partition assignment.

        :param partition_id: (int), the partition id.
        :return: (:class:`~hazelcast.core.Address`), owner of partition or ``None`` if it's not set yet.
        """
        if partition_id not in self.partitions:
            self._do_refresh()
        return self.partitions.get(partition_id, None)

    def get_partition_id(self, key):
        """
        Returns the partition id for a Data key.

        :param key: (object), the data key.
        :return: (int), the partition id.
        """
        data = self._client.serialization_service.to_data(key)
        count = self.get_partition_count()
        if count <= 0:
            return 0
        return hash_to_index(data.get_partition_hash(), count)

    def get_partition_count(self):
        """
        Returns the number of partitions.

        :return: (int), the number of partitions.
        """
        if not self.partitions:
            self._get_partition_count_blocking()
        return len(self.partitions)

    def _get_partition_count_blocking(self):
        event = threading.Event()
        while not event.isSet():
            self._do_refresh(callback=lambda: event.set())
            event.wait(timeout=1)

    def _do_refresh(self, callback=None):
        self.logger.debug("Start updating partitions", extra=self._logger_extras)
        address = self._client.cluster.owner_connection_address
        connection = self._client.connection_manager.get_connection(address)
        if connection is None:
            self.logger.debug("Could not update partition thread as owner connection is not available.",
                              extra=self._logger_extras)
            if callback:
                callback()
            return
        request = client_get_partitions_codec.encode_request()

        def cb(f):
            if f.is_success():
                self.process_partition_response(f.result())
                if callback:
                    callback()

        future = self._client.invoker.invoke_on_connection(request, connection)
        future.add_done_callback(cb)

    def process_partition_response(self, message):
        partitions = client_get_partitions_codec.decode_response(message)["partitions"]
        partitions_dict = {}
        for addr, partition_list in six.iteritems(partitions):
            for partition in partition_list:
                partitions_dict[partition] = addr
        self.partitions.update(partitions_dict)
        self.logger.debug("Finished updating partitions", extra=self._logger_extras)


def string_partition_strategy(key):
    if key is None:
        return None
    try:
        index_of = key.index('@')
        return key[index_of + 1:]
    except ValueError:
        return key



"""
Event Response Constants
"""

EVENT_MEMBER = 200
EVENT_MEMBERLIST = 201
EVENT_MEMBERATTRIBUTECHANGE = 202
EVENT_ENTRY = 203
EVENT_ITEM = 204
EVENT_TOPIC = 205
EVENT_PARTITIONLOST = 206
EVENT_DISTRIBUTEDOBJECT = 207
EVENT_CACHEINVALIDATION = 208
EVENT_MAPPARTITIONLOST = 209
EVENT_CACHE = 210
EVENT_CACHEBATCHINVALIDATION = 211
# ENTERPRISE
EVENT_QUERYCACHESINGLE = 212
EVENT_QUERYCACHEBATCH = 213

EVENT_CACHEPARTITIONLOST = 214
EVENT_IMAPINVALIDATION = 215
EVENT_IMAPBATCHINVALIDATION = 216



from hazelcast.protocol.codec import lock_force_unlock_codec, lock_get_lock_count_codec, \
    lock_get_remaining_lease_time_codec, lock_is_locked_by_current_thread_codec, lock_is_locked_codec, lock_lock_codec, \
    lock_try_lock_codec, lock_unlock_codec
from hazelcast.proxy.base import PartitionSpecificProxy, MAX_SIZE
from hazelcast.util import thread_id, to_millis


class Lock(PartitionSpecificProxy):
    """
    Distributed implementation of `asyncio.Lock <https://docs.python.org/3/library/asyncio-sync.html>`_
    """

    def __init__(self, client, service_name, name):
        super(Lock, self).__init__(client, service_name, name)
        self.reference_id_generator = self._client.lock_reference_id_generator

    def force_unlock(self):
        """
        Releases the lock regardless of the lock owner. It always successfully unlocks, never blocks, and returns
        immediately.
        """
        return self._encode_invoke(lock_force_unlock_codec,
                                   reference_id=self.reference_id_generator.get_and_increment())

    def get_lock_count(self):
        """
        Returns re-entrant lock hold count, regardless of lock ownership.

        :return: (int), the lock hold count.
        """
        return self._encode_invoke(lock_get_lock_count_codec)

    def get_remaining_lease_time(self):
        """
        Returns remaining lease time in milliseconds. If the lock is not locked then -1 will be returned.

        :return: (long), remaining lease time in milliseconds.
        """
        return self._encode_invoke(lock_get_remaining_lease_time_codec)

    def is_locked(self):
        """
        Returns whether this lock is locked or not.

        :return: (bool), ``true`` if this lock is locked, ``false`` otherwise.
        """
        return self._encode_invoke(lock_is_locked_codec)

    def is_locked_by_current_thread(self):
        """
        Returns whether this lock is locked by current thread or not.

        :return: (bool), ``true`` if this lock is locked by current thread, ``false`` otherwise.
        """
        return self._encode_invoke(lock_is_locked_by_current_thread_codec,
                                   thread_id=thread_id())

    def lock(self, lease_time=-1):
        """
        Acquires the lock. If a lease time is specified, lock will be released after this lease time.

        If the lock is not available, the current thread becomes disabled for thread scheduling purposes and lies
        dormant until the lock has been acquired.

        :param lease_time: (long), time to wait before releasing the lock (optional).
        """
        return self._encode_invoke(lock_lock_codec, invocation_timeout=MAX_SIZE, lease_time=to_millis(lease_time),
                                   thread_id=thread_id(), reference_id=self.reference_id_generator.get_and_increment())

    def try_lock(self, timeout=0, lease_time=-1):
        """
        Tries to acquire the lock. When the lock is not available,

            * If timeout is not provided, the current thread doesn't wait and returns ``false`` immediately.
            * If a timeout is provided, the current thread becomes disabled for thread scheduling purposes and lies
            dormant until one of the followings happens:
                * the lock is acquired by the current thread, or
                * the specified waiting time elapses.

        If lease time is provided, lock will be released after this time elapses.

        :param timeout: (long), maximum time in seconds to wait for the lock (optional).
        :param lease_time: (long), time in seconds to wait before releasing the lock (optional).
        :return: (bool), ``true`` if the lock was acquired and otherwise, ``false``.
        """
        return self._encode_invoke(lock_try_lock_codec, invocation_timeout=MAX_SIZE, lease=to_millis(lease_time),
                                   thread_id=thread_id(), timeout=to_millis(timeout),
                                   reference_id=self.reference_id_generator.get_and_increment())

    def unlock(self):
        """
        Releases the lock.
        """
        return self._encode_invoke(lock_unlock_codec, thread_id=thread_id(),
                                   reference_id=self.reference_id_generator.get_and_increment())



import logging
import struct
import sys
import threading
import time

from hazelcast.exception import AuthenticationError
from hazelcast.future import ImmediateFuture, ImmediateExceptionFuture
from hazelcast.protocol.client_message import BEGIN_END_FLAG, ClientMessage, ClientMessageBuilder
from hazelcast.protocol.codec import client_authentication_codec, client_ping_codec
from hazelcast.serialization import INT_SIZE_IN_BYTES, FMT_LE_INT
from hazelcast.util import AtomicInteger, parse_addresses, calculate_version
from hazelcast.version import CLIENT_TYPE, CLIENT_VERSION, SERIALIZATION_VERSION
from hazelcast import six

BUFFER_SIZE = 128000
PROTOCOL_VERSION = 1


class ConnectionManager(object):
    """
    ConnectionManager is responsible for managing :mod:`Connection` objects.
    """
    logger = logging.getLogger("HazelcastClient.ConnectionManager")

    def __init__(self, client, new_connection_func, address_translator):
        self._new_connection_mutex = threading.RLock()
        self._io_thread = None
        self._client = client
        self.connections = {}
        self._pending_connections = {}
        self._socket_map = {}
        self._new_connection_func = new_connection_func
        self._connection_listeners = []
        self._address_translator = address_translator
        self._logger_extras = {"client_name": client.name, "group_name": client.config.group_config.name}

    def add_listener(self, on_connection_opened=None, on_connection_closed=None):
        """
        Registers a ConnectionListener. If the same listener is registered multiple times, it will be notified multiple
        times.

        :param on_connection_opened: (Function), function to be called when a connection is opened.
        :param on_connection_closed: (Function), function to be called when a connection is removed.
        """
        self._connection_listeners.append((on_connection_opened, on_connection_closed))

    def get_connection(self, address):
        """
        Gets the existing connection for a given address or connects. This call is silent.

        :param address: (:class:`~hazelcast.core.Address`), the address to connect to.
        :return: (:class:`~hazelcast.connection.Connection`), the found connection, or None if no connection exists.
        """
        try:
            return self.connections[address]
        except KeyError:
            return None

    def _cluster_authenticator(self, connection):
        uuid = self._client.cluster.uuid
        owner_uuid = self._client.cluster.owner_uuid

        request = client_authentication_codec.encode_request(
            username=self._client.config.group_config.name,
            password=self._client.config.group_config.password,
            uuid=uuid,
            owner_uuid=owner_uuid,
            is_owner_connection=False,
            client_type=CLIENT_TYPE,
            serialization_version=SERIALIZATION_VERSION,
            client_hazelcast_version=CLIENT_VERSION)

        def callback(f):
            parameters = client_authentication_codec.decode_response(f.result())
            if parameters["status"] != 0:
                raise AuthenticationError("Authentication failed.")
            connection.endpoint = parameters["address"]
            self.owner_uuid = parameters["owner_uuid"]
            self.uuid = parameters["uuid"]
            connection.server_version_str = parameters.get("server_hazelcast_version", "")
            connection.server_version = calculate_version(connection.server_version_str)
            return connection

        return self._client.invoker.invoke_on_connection(request, connection).continue_with(callback)

    def get_or_connect(self, address, authenticator=None):
        """
        Gets the existing connection for a given address. If it does not exist, the system will try to connect
        asynchronously. In this case, it returns a Future. When the connection is established at some point in time, it
        can be retrieved by using the get_connection(:class:`~hazelcast.core.Address`) or from Future.

        :param address: (:class:`~hazelcast.core.Address`), the address to connect to.
        :param authenticator: (Function), function to be used for authentication (optional).
        :return: (:class:`~hazelcast.connection.Connection`), the existing connection or it returns a Future which includes asynchronously.
        """
        if address in self.connections:
            return ImmediateFuture(self.connections[address])
        else:
            with self._new_connection_mutex:
                if address in self._pending_connections:
                    return self._pending_connections[address]
                else:
                    authenticator = authenticator or self._cluster_authenticator
                    try:
                        translated_address = self._address_translator.translate(address)
                        if translated_address is None:
                            raise ValueError("Address translator could not translate address: {}".format(address))
                        connection = self._new_connection_func(translated_address,
                                                               self._client.config.network_config.connection_timeout,
                                                               self._client.config.network_config.socket_options,
                                                               connection_closed_callback=self._connection_closed,
                                                               message_callback=self._client.invoker._handle_client_message,
                                                               network_config=self._client.config.network_config)
                    except IOError:
                        return ImmediateExceptionFuture(sys.exc_info()[1], sys.exc_info()[2])

                    future = authenticator(connection).continue_with(self.on_auth, connection, address)
                    if not future.done():
                        self._pending_connections[address] = future
                    return future

    def on_auth(self, f, connection, address):
        """
        Checks for authentication of a connection.

        :param f: (:class:`~hazelcast.future.Future`), future that contains the result of authentication.
        :param connection: (:class:`~hazelcast.connection.Connection`), newly established connection.
        :param address: (:class:`~hazelcast.core.Address`), the adress of new connection.
        :return: Result of authentication.
        """
        if f.is_success():
            self.logger.info("Authenticated with %s", f.result(), extra=self._logger_extras)
            with self._new_connection_mutex:
                self.connections[connection.endpoint] = f.result()
                try:
                    self._pending_connections.pop(address)
                except KeyError:
                    pass
            for on_connection_opened, _ in self._connection_listeners:
                if on_connection_opened:
                    on_connection_opened(f.result())
            return f.result()
        else:
            self.logger.debug("Error opening %s", connection, extra=self._logger_extras)
            with self._new_connection_mutex:
                try:
                    self._pending_connections.pop(address)
                except KeyError:
                    pass
            six.reraise(f.exception().__class__, f.exception(), f.traceback())

    def _connection_closed(self, connection, cause):
        # if connection was authenticated, fire event
        if connection.endpoint:
            try:
                self.connections.pop(connection.endpoint)
            except KeyError:
                pass
            for _, on_connection_closed in self._connection_listeners:
                if on_connection_closed:
                    on_connection_closed(connection, cause)
        else:
            # clean-up unauthenticated connection
            self._client.invoker.cleanup_connection(connection, cause)

    def close_connection(self, address, cause):
        """
        Closes the connection with given address.

        :param address: (:class:`~hazelcast.core.Address`), address of the connection to be closed.
        :param cause: (Exception), the cause for closing the connection.
        :return: (bool), ``true`` if the connection is closed, ``false`` otherwise.
        """
        try:
            connection = self.connections[address]
            connection.close(cause)
        except KeyError:
            self.logger.warning("No connection with %s was found to close.", address, extra=self._logger_extras)
            return False


class Heartbeat(object):
    """
    HeartBeat Service.
    """
    _heartbeat_timer = None
    logger = logging.getLogger("HazelcastClient.HeartbeatService")

    def __init__(self, client):
        self._client = client
        self._listeners = []
        self._logger_extras = {"client_name": client.name, "group_name": client.config.group_config.name}

        self._heartbeat_timeout = client.properties.get_seconds_positive_or_default(client.properties.HEARTBEAT_TIMEOUT)
        self._heartbeat_interval = client.properties.get_seconds_positive_or_default(client.properties.HEARTBEAT_INTERVAL)

    def start(self):
        """
        Starts sending periodic HeartBeat operations.
        """
        def _heartbeat():
            if not self._client.lifecycle.is_live:
                return
            self._heartbeat()
            self._heartbeat_timer = self._client.reactor.add_timer(self._heartbeat_interval, _heartbeat)

        self._heartbeat_timer = self._client.reactor.add_timer(self._heartbeat_interval, _heartbeat)

    def shutdown(self):
        """
        Stops HeartBeat operations.
        """
        if self._heartbeat_timer:
            self._heartbeat_timer.cancel()

    def add_listener(self, on_heartbeat_restored=None, on_heartbeat_stopped=None):
        """
        Registers a HeartBeat listener. Listener is invoked when a HeartBeat related event occurs.

        :param on_heartbeat_restored: (Function), function to be called when a HeartBeat is restored (optional).
        :param on_heartbeat_stopped:  (Function), function to be called when a HeartBeat is stopped (optional).
        """
        self._listeners.append((on_heartbeat_restored, on_heartbeat_stopped))

    def _heartbeat(self):
        now = time.time()
        for connection in list(self._client.connection_manager.connections.values()):
            time_since_last_read = now - connection.last_read_in_seconds
            time_since_last_write = now - connection.last_write_in_seconds
            if time_since_last_read > self._heartbeat_timeout:
                if connection.heartbeating:
                    self.logger.warning(
                        "Heartbeat: Did not hear back after %ss from %s" % (time_since_last_read, connection),
                        extra=self._logger_extras)
                    self._on_heartbeat_stopped(connection)
            else:
                if not connection.heartbeating:
                    self._on_heartbeat_restored(connection)

            if time_since_last_write > self._heartbeat_interval:
                request = client_ping_codec.encode_request()
                self._client.invoker.invoke_on_connection(request, connection, ignore_heartbeat=True)

    def _on_heartbeat_restored(self, connection):
        self.logger.info("Heartbeat: Heartbeat restored for connection %s" % connection, extra=self._logger_extras)
        connection.heartbeating = True
        for callback, _ in self._listeners:
            if callback:
                callback(connection)

    def _on_heartbeat_stopped(self, connection):
        connection.heartbeating = False
        for _, callback in self._listeners:
            if callback:
                callback(connection)


class Connection(object):
    """
    Connection object which stores connection related information and operations.
    """
    _closed = False
    endpoint = None
    heartbeating = True
    is_owner = False
    counter = AtomicInteger()

    def __init__(self, address, connection_closed_callback, message_callback, logger_extras=None):
        self._address = (address.host, address.port)
        self._logger_extras = logger_extras
        self.id = self.counter.get_and_increment()
        self.logger = logging.getLogger("HazelcastClient.Connection[%s](%s:%d)" % (self.id, address.host, address.port))
        self._connection_closed_callback = connection_closed_callback
        self._builder = ClientMessageBuilder(message_callback)
        self._read_buffer = bytearray()
        self.last_read_in_seconds = 0
        self.last_write_in_seconds = 0
        self.start_time_in_seconds = 0
        self.server_version_str = ""
        self.server_version = 0

    def live(self):
        """
        Determines whether this connection is live or not.

        :return: (bool), ``true`` if the connection is live, ``false`` otherwise.
        """
        return not self._closed

    def send_message(self, message):
        """
        Sends a message to this connection.

        :param message: (Message), message to be sent to this connection.
        """
        if not self.live():
            raise IOError("Connection is not live.")

        message.add_flag(BEGIN_END_FLAG)
        self.write(message.buffer)

    def receive_message(self):
        """
        Receives a message from this connection.
        """
        # split frames
        while len(self._read_buffer) >= INT_SIZE_IN_BYTES:
            frame_length = struct.unpack_from(FMT_LE_INT, self._read_buffer, 0)[0]
            if frame_length > len(self._read_buffer):
                return
            message = ClientMessage(memoryview(self._read_buffer)[:frame_length])
            self._read_buffer = self._read_buffer[frame_length:]
            self._builder.on_message(message)

    def write(self, data):
        """
        Writes data to this connection when sending messages.

        :param data: (Data), data to be written to connection.
        """
        # must be implemented by subclass
        raise NotImplementedError

    def close(self, cause):
        """
        Closes the connection.

        :param cause: (Exception), the cause of closing the connection.
        """
        raise NotImplementedError

    def __repr__(self):
        return "Connection(address=%s, id=%s)" % (self._address, self.id)

    def __hash__(self):
        return self.id


class DefaultAddressProvider(object):
    """
    Provides initial addresses for client to find and connect to a node.
    Loads addresses from the Hazelcast configuration.
    """
    def __init__(self, network_config):
        self._network_config = network_config

    def load_addresses(self):
        """
        :return: (Sequence), The possible member addresses to connect to.
        """
        return parse_addresses(self._network_config.addresses)


class DefaultAddressTranslator(object):
    """
    DefaultAddressTranslator is a no-op. It always returns the given address.
    """
    def translate(self, address):
        """
        :param address: (:class:`~hazelcast.core.Address`), address to be translated.
        :return: (:class:`~hazelcast.core.Address`), translated address.
        """
        return address

    def refresh(self):
        """Refreshes the internal lookup table if necessary."""
        pass



import logging
import logging.config
import sys
import json

from hazelcast.cluster import ClusterService, RandomLoadBalancer
from hazelcast.config import ClientConfig, ClientProperties
from hazelcast.connection import ConnectionManager, Heartbeat, DefaultAddressProvider, DefaultAddressTranslator
from hazelcast.invocation import InvocationService
from hazelcast.listener import ListenerService
from hazelcast.lifecycle import LifecycleService, LIFECYCLE_STATE_SHUTTING_DOWN, LIFECYCLE_STATE_SHUTDOWN
from hazelcast.partition import PartitionService
from hazelcast.proxy import ProxyManager, MAP_SERVICE, QUEUE_SERVICE, LIST_SERVICE, SET_SERVICE, MULTI_MAP_SERVICE, \
    REPLICATED_MAP_SERVICE, ATOMIC_LONG_SERVICE, ATOMIC_REFERENCE_SERVICE, RINGBUFFER_SERIVCE, COUNT_DOWN_LATCH_SERVICE, \
    TOPIC_SERVICE, RELIABLE_TOPIC_SERVICE, SEMAPHORE_SERVICE, LOCK_SERVICE, ID_GENERATOR_SERVICE, \
    ID_GENERATOR_ATOMIC_LONG_PREFIX, EXECUTOR_SERVICE, PN_COUNTER_SERVICE, FLAKE_ID_GENERATOR_SERVICE
from hazelcast.near_cache import NearCacheManager
from hazelcast.reactor import AsyncoreReactor
from hazelcast.serialization import SerializationServiceV1
from hazelcast.statistics import Statistics
from hazelcast.transaction import TWO_PHASE, TransactionManager
from hazelcast.util import AtomicInteger, DEFAULT_LOGGING
from hazelcast.discovery import HazelcastCloudAddressProvider, HazelcastCloudAddressTranslator, HazelcastCloudDiscovery
from hazelcast.exception import HazelcastIllegalStateError


class HazelcastClient(object):
    """
    Hazelcast Client.
    """
    CLIENT_ID = AtomicInteger()
    _config = None
    logger = logging.getLogger("HazelcastClient")

    def __init__(self, config=None):
        self.config = config or ClientConfig()
        self.properties = ClientProperties(self.config.get_properties())
        self.id = HazelcastClient.CLIENT_ID.get_and_increment()
        self.name = self._create_client_name()
        self._init_logger()
        self._logger_extras = {"client_name": self.name, "group_name": self.config.group_config.name}
        self._log_group_password_info()
        self.lifecycle = LifecycleService(self.config, self._logger_extras)
        self.reactor = AsyncoreReactor(self._logger_extras)
        self._address_providers = self._create_address_providers()
        self._address_translator = self._create_address_translator()
        self.connection_manager = ConnectionManager(self, self.reactor.new_connection, self._address_translator)
        self.heartbeat = Heartbeat(self)
        self.invoker = InvocationService(self)
        self.listener = ListenerService(self)
        self.cluster = ClusterService(self.config, self, self._address_providers)
        self.partition_service = PartitionService(self)
        self.proxy = ProxyManager(self)
        self.load_balancer = RandomLoadBalancer(self.cluster)
        self.serialization_service = SerializationServiceV1(serialization_config=self.config.serialization_config, properties=self.properties)
        self.transaction_manager = TransactionManager(self)
        self.lock_reference_id_generator = AtomicInteger(1)
        self.near_cache_manager = NearCacheManager(self)
        self.statistics = Statistics(self)
        self._start()

    def _start(self):
        self.reactor.start()
        try:
            self.invoker.start()
            self.cluster.start()
            self.heartbeat.start()
            self.listener.start()
            self.partition_service.start()
            self.statistics.start()
        except:
            self.reactor.shutdown()
            raise
        self.logger.info("Client started.", extra=self._logger_extras)

    def get_atomic_long(self, name):
        """
        Creates cluster-wide :class:`~hazelcast.proxy.atomic_long.AtomicLong`.

        :param name: (str), name of the AtomicLong proxy.
        :return: (:class:`~hazelcast.proxy.atomic_long.AtomicLong`), AtomicLong proxy for the given name.
        """
        return self.proxy.get_or_create(ATOMIC_LONG_SERVICE, name)

    def get_atomic_reference(self, name):
        """
        Creates cluster-wide :class:`~hazelcast.proxy.atomic_reference.AtomicReference`.

        :param name: (str), name of the AtomicReference proxy.
        :return: (:class:`~hazelcast.proxy.atomic_reference.AtomicReference`), AtomicReference proxy for the given name.
        """
        return self.proxy.get_or_create(ATOMIC_REFERENCE_SERVICE, name)

    def get_count_down_latch(self, name):
        """
        Creates cluster-wide :class:`~hazelcast.proxy.count_down_latch.CountDownLatch`.

        :param name: (str), name of the CountDownLatch proxy.
        :return: (:class:`~hazelcast.proxy.count_down_latch.CountDownLatch`), CountDownLatch proxy for the given name.
        """
        return self.proxy.get_or_create(COUNT_DOWN_LATCH_SERVICE, name)

    def get_executor(self, name):
        """
        Creates cluster-wide :class:`~hazelcast.proxy.executor.Executor`.

        :param name: (str), name of the Executor proxy.
        :return: (:class:`~hazelcast.proxy.executor.Executor`), Executor proxy for the given name.
        """
        return self.proxy.get_or_create(EXECUTOR_SERVICE, name)

    def get_flake_id_generator(self, name):
        """
        Creates or returns a cluster-wide :class:`~hazelcast.proxy.flake_id_generator.FlakeIdGenerator`.

        :param name: (str), name of the FlakeIdGenerator proxy.
        :return: (:class:`~hazelcast.proxy.flake_id_generator.FlakeIdGenerator`), FlakeIdGenerator proxy for the given name
        """
        return self.proxy.get_or_create(FLAKE_ID_GENERATOR_SERVICE, name)

    def get_id_generator(self, name):
        """
        Creates cluster-wide :class:`~hazelcast.proxy.id_generator.IdGenerator`.

        :param name: (str), name of the IdGenerator proxy.
        :return: (:class:`~hazelcast.proxy.id_generator.IdGenerator`), IdGenerator proxy for the given name.
        """
        atomic_long = self.get_atomic_long(ID_GENERATOR_ATOMIC_LONG_PREFIX + name)
        return self.proxy.get_or_create(ID_GENERATOR_SERVICE, name, atomic_long=atomic_long)

    def get_queue(self, name):
        """
        Returns the distributed queue instance with the specified name.

        :param name: (str), name of the distributed queue.
        :return: (:class:`~hazelcast.proxy.queue.Queue`), distributed queue instance with the specified name.
        """
        return self.proxy.get_or_create(QUEUE_SERVICE, name)

    def get_list(self, name):
        """
        Returns the distributed list instance with the specified name.

        :param name: (str), name of the distributed list.
        :return: (:class:`~hazelcast.proxy.list.List`), distributed list instance with the specified name.
        """
        return self.proxy.get_or_create(LIST_SERVICE, name)

    def get_lock(self, name):
        """
        Returns the distributed lock instance with the specified name.

        :param name: (str), name of the distributed lock.
        :return: (:class:`~hazelcast.proxy.lock.Lock`), distributed lock instance with the specified name.
        """
        return self.proxy.get_or_create(LOCK_SERVICE, name)

    def get_map(self, name):
        """
        Returns the distributed map instance with the specified name.

        :param name: (str), name of the distributed map.
        :return: (:class:`~hazelcast.proxy.map.Map`), distributed map instance with the specified name.
        """
        return self.proxy.get_or_create(MAP_SERVICE, name)

    def get_multi_map(self, name):
        """
        Returns the distributed MultiMap instance with the specified name.

        :param name: (str), name of the distributed MultiMap.
        :return: (:class:`~hazelcast.proxy.multi_map.MultiMap`), distributed MultiMap instance with the specified name.
        """
        return self.proxy.get_or_create(MULTI_MAP_SERVICE, name)

    def get_pn_counter(self, name):
        """
        Returns the PN Counter instance with the specified name.

        :param name: (str), name of the PN Counter.
        :return: (:class:`~hazelcast.proxy.pn_counter.PNCounter`), the PN Counter.
        """
        return self.proxy.get_or_create(PN_COUNTER_SERVICE, name)

    def get_reliable_topic(self, name):
        """
        Returns the :class:`~hazelcast.proxy.reliable_topic.ReliableTopic` instance with the specified name.

        :param name: (str), name of the ReliableTopic.
        :return: (:class:`~hazelcast.proxy.reliable_topic.ReliableTopic`), the ReliableTopic.
        """
        return self.proxy.get_or_create(RELIABLE_TOPIC_SERVICE, name)

    def get_replicated_map(self, name):
        """
        Returns the distributed ReplicatedMap instance with the specified name.

        :param name: (str), name of the distributed ReplicatedMap.
        :return: (:class:`~hazelcast.proxy.replicated_map.ReplicatedMap`), distributed ReplicatedMap instance with the specified name.
        """
        return self.proxy.get_or_create(REPLICATED_MAP_SERVICE, name)

    def get_ringbuffer(self, name):
        """
        Returns the distributed RingBuffer instance with the specified name.

        :param name: (str), name of the distributed RingBuffer.
        :return: (:class:`~hazelcast.proxy.ringbuffer.RingBuffer`), distributed RingBuffer instance with the specified name.
        """

        return self.proxy.get_or_create(RINGBUFFER_SERIVCE, name)

    def get_semaphore(self, name):
        """
        Returns the distributed Semaphore instance with the specified name.

        :param name: (str), name of the distributed Semaphore.
        :return: (:class:`~hazelcast.proxy.semaphore.Semaphore`), distributed Semaphore instance with the specified name.
        """
        return self.proxy.get_or_create(SEMAPHORE_SERVICE, name)

    def get_set(self, name):
        """
        Returns the distributed Set instance with the specified name.

        :param name: (str), name of the distributed Set.
        :return: (:class:`~hazelcast.proxy.set.Set`), distributed Set instance with the specified name.
        """
        return self.proxy.get_or_create(SET_SERVICE, name)

    def get_topic(self, name):
        """
        Returns the :class:`~hazelcast.proxy.topic.Topic` instance with the specified name.

        :param name: (str), name of the Topic.
        :return: (:class:`~hazelcast.proxy.topic.Topic`), the Topic.
        """
        return self.proxy.get_or_create(TOPIC_SERVICE, name)

    def new_transaction(self, timeout=120, durability=1, type=TWO_PHASE):
        """
        Creates a new :class:`~hazelcast.transaction.Transaction` associated with the current thread using default or given options.

        :param timeout: (long), the timeout in seconds determines the maximum lifespan of a transaction. So if a
            transaction is configured with a timeout of 2 minutes, then it will automatically rollback if it hasn't
            committed yet.
        :param durability: (int), the durability is the number of machines that can take over if a member fails during a
        transaction commit or rollback
        :param type: (Transaction Type), the transaction type which can be :const:`~hazelcast.transaction.TWO_PHASE` or :const:`~hazelcast.transaction.ONE_PHASE`
        :return: (:class:`~hazelcast.transaction.Transaction`), new Transaction associated with the current thread.
        """
        return self.transaction_manager.new_transaction(timeout, durability, type)

    def shutdown(self):
        """
        Shuts down this HazelcastClient.
        """
        if self.lifecycle.is_live:
            self.lifecycle.fire_lifecycle_event(LIFECYCLE_STATE_SHUTTING_DOWN)
            self.near_cache_manager.destroy_all_near_caches()
            self.statistics.shutdown()
            self.partition_service.shutdown()
            self.heartbeat.shutdown()
            self.cluster.shutdown()
            self.reactor.shutdown()
            self.lifecycle.fire_lifecycle_event(LIFECYCLE_STATE_SHUTDOWN)
            self.logger.info("Client shutdown.", extra=self._logger_extras)

    def _create_address_providers(self):
        network_config = self.config.network_config
        address_providers = []

        cloud_config = network_config.cloud_config
        cloud_address_provider = self._init_cloud_address_provider(cloud_config)
        if cloud_address_provider:
            address_providers.append(cloud_address_provider)

        address_providers.append(DefaultAddressProvider(network_config))
        return address_providers

    def _init_cloud_address_provider(self, cloud_config):
        if cloud_config.enabled:
            discovery_token = cloud_config.discovery_token
            host, url = HazelcastCloudDiscovery.get_host_and_url(self.config.get_properties(), discovery_token)
            return HazelcastCloudAddressProvider(host, url, self._get_connection_timeout(), self._logger_extras)

        cloud_token = self.properties.get(self.properties.HAZELCAST_CLOUD_DISCOVERY_TOKEN)
        if cloud_token != "":
            host, url = HazelcastCloudDiscovery.get_host_and_url(self.config.get_properties(), cloud_token)
            return HazelcastCloudAddressProvider(host, url, self._get_connection_timeout(), self._logger_extras)

        return None

    def _create_address_translator(self):
        network_config = self.config.network_config
        cloud_config = network_config.cloud_config
        cloud_discovery_token = self.properties.get(self.properties.HAZELCAST_CLOUD_DISCOVERY_TOKEN)

        address_list_provided = len(network_config.addresses) != 0
        if cloud_discovery_token != "" and cloud_config.enabled:
            raise HazelcastIllegalStateError("Ambiguous Hazelcast.cloud configuration. "
                                             "Both property based and client configuration based settings are provided "
                                             "for Hazelcast cloud discovery together. Use only one.")

        hazelcast_cloud_enabled = cloud_discovery_token != "" or cloud_config.enabled
        self._is_discovery_configuration_consistent(address_list_provided, hazelcast_cloud_enabled)

        if hazelcast_cloud_enabled:
            if cloud_config.enabled:
                discovery_token = cloud_config.discovery_token
            else:
                discovery_token = cloud_discovery_token
            host, url = HazelcastCloudDiscovery.get_host_and_url(self.config.get_properties(), discovery_token)
            return HazelcastCloudAddressTranslator(host, url, self._get_connection_timeout(), self._logger_extras)

        return DefaultAddressTranslator()

    def _get_connection_timeout(self):
        network_config = self.config.network_config
        conn_timeout = network_config.connection_timeout
        return sys.maxsize if conn_timeout == 0 else conn_timeout

    def _is_discovery_configuration_consistent(self, address_list_provided, hazelcast_cloud_enabled):
        count = 0
        if address_list_provided:
            count += 1
        if hazelcast_cloud_enabled:
            count += 1

        if count > 1:
            raise HazelcastIllegalStateError("Only one discovery method can be enabled at a time. "
                                             "Cluster members given explicitly: {}"
                                             ", Hazelcast.cloud enabled: {}".format(address_list_provided,
                                                                                    hazelcast_cloud_enabled))

    def _create_client_name(self):
        if self.config.client_name:
            return self.config.client_name
        return "hz.client_" + str(self.id)

    def _init_logger(self):
        logger_config = self.config.logger_config
        if logger_config.config_file is not None:
            with open(logger_config.config_file, "r") as f:
                json_config = json.loads(f.read())
                logging.config.dictConfig(json_config)
        else:
            logging.config.dictConfig(DEFAULT_LOGGING)
            self.logger.setLevel(logger_config.level)

    def _log_group_password_info(self):
        if self.config.group_config.password:
            self.logger.info("A non-empty group password is configured for the Hazelcast client. "
                             "Starting with Hazelcast IMDG version 3.11, clients with the same group name, "
                             "but with different group passwords (that do not use authentication) will be "
                             "accepted to a cluster. The group password configuration will be removed "
                             "completely in a future release.", extra=self._logger_extras)




SEMAPHORE_INIT = 0x0d01
SEMAPHORE_ACQUIRE = 0x0d02
SEMAPHORE_AVAILABLEPERMITS = 0x0d03
SEMAPHORE_DRAINPERMITS = 0x0d04
SEMAPHORE_REDUCEPERMITS = 0x0d05
SEMAPHORE_RELEASE = 0x0d06
SEMAPHORE_TRYACQUIRE = 0x0d07



from hazelcast.protocol.codec import transactional_multi_map_get_codec, transactional_multi_map_put_codec, \
    transactional_multi_map_remove_codec, transactional_multi_map_remove_entry_codec, \
    transactional_multi_map_size_codec, transactional_multi_map_value_count_codec
from hazelcast.proxy.base import TransactionalProxy
from hazelcast.util import check_not_none


class TransactionalMultiMap(TransactionalProxy):
    """
    Transactional implementation of :class:`~hazelcast.proxy.multi_map.MultiMap`.
    """
    def put(self, key, value):
        """
        Transactional implementation of :func:`MultiMap.put(key, value) <hazelcast.proxy.multi_map.MultiMap.put>`

        :param key: (object), the key to be stored.
        :param value: (object), the value to be stored.
        :return: (bool), ``true`` if the size of the multimap is increased, ``false`` if the multimap already contains the
        key-value tuple.
        """
        check_not_none(key, "key can't be none")
        check_not_none(value, "value can't be none")
        return self._encode_invoke(transactional_multi_map_put_codec, key=self._to_data(key),
                                   value=self._to_data(value))

    def get(self, key):
        """
        Transactional implementation of :func:`MultiMap.get(key) <hazelcast.proxy.multi_map.MultiMap.get>`

        :param key: (object), the key whose associated values are returned.
        :return: (Sequence), the collection of the values associated with the key.
        """
        check_not_none(key, "key can't be none")
        return self._encode_invoke(transactional_multi_map_get_codec, key=self._to_data(key))

    def remove(self, key, value):
        """
        Transactional implementation of :func:`MultiMap.remove(key, value)
        <hazelcast.proxy.multi_map.MultiMap.remove>`

        :param key: (object), the key of the entry to remove.
        :param value: (object), the value of the entry to remove.
        :return:
        """
        check_not_none(key, "key can't be none")
        check_not_none(value, "value can't be none")
        return self._encode_invoke(transactional_multi_map_remove_entry_codec, key=self._to_data(key),
                                   value=self._to_data(value))

    def remove_all(self, key):
        """
        Transactional implementation of :func:`MultiMap.remove_all(key)
        <hazelcast.proxy.multi_map.MultiMap.remove_all>`

        :param key: (object), the key of the entries to remove.
        :return: (list), the collection of the values associated with the key.
        """
        check_not_none(key, "key can't be none")
        return self._encode_invoke(transactional_multi_map_remove_codec, key=self._to_data(key))

    def value_count(self, key):
        """
        Transactional implementation of :func:`MultiMap.value_count(key)
        <hazelcast.proxy.multi_map.MultiMap.value_count>`

        :param key: (object), the key whose number of values is to be returned.
        :return: (int), the number of values matching the given key in the multimap.
        """
        check_not_none(key, "key can't be none")
        return self._encode_invoke(transactional_multi_map_value_count_codec, key=self._to_data(key))

    def size(self):
        """
        Transactional implementation of :func:`MultiMap.size() <hazelcast.proxy.multi_map.MultiMap.size>`

        :return: (int), the number of key-value tuples in the multimap.
        """
        return self._encode_invoke(transactional_multi_map_size_codec)



import logging
import sys
import threading

from hazelcast.util import AtomicInteger
from hazelcast import six

NONE_RESULT = object()


class Future(object):
    """
    Future is used for representing an asynchronous computation result.
    """
    _result = None
    _exception = None
    _traceback = None
    _threading_locals = threading.local()
    logger = logging.getLogger("HazelcastClient.Future")

    def __init__(self):
        self._callbacks = []
        self._event = _Event()

    def set_result(self, result):
        """
        Sets the result of the Future.

        :param result: Result of the Future.
        """
        if result is None:
            self._result = NONE_RESULT
        else:
            self._result = result
        self._event.set()
        self._invoke_callbacks()

    def set_exception(self, exception, traceback=None):
        """
        Sets the exception for this Future in case of errors.

        :param exception: (Exception), exception to be threw in case of error.
        :param traceback: (Function), function to be called on traceback (optional).
        """
        if not isinstance(exception, BaseException):
            raise RuntimeError("Exception must be of BaseException type")
        self._exception = exception
        self._traceback = traceback
        self._event.set()
        self._invoke_callbacks()

    def result(self):
        """
        Returns the result of the Future, which makes the call synchronous if the result has not been computed yet.

        :return: Result of the Future.
        """
        self._reactor_check()
        self._event.wait()
        if self._exception:
            six.reraise(self._exception.__class__, self._exception, self._traceback)
        if self._result == NONE_RESULT:
            return None
        else:
            return self._result

    def _reactor_check(self):
        if not self.done() and hasattr(self._threading_locals, 'is_reactor_thread'):
            raise RuntimeError(
                    "Synchronous result for incomplete operation must not be called from Reactor thread. "
                    "Use add_done_callback instead.")

    def is_success(self):
        """
        Determines whether the result can be successfully computed or not.
        """
        return self._result is not None

    def done(self):
        """
        Determines whether the result is computed or not.

        :return: (bool), ``true`` if the result is computed, ``false`` otherwise.
        """
        return self._event.is_set()

    def running(self):
        """
        Determines whether the asynchronous call, the computation is still running or not.

        :return: (bool), ``true`` if the  result is being computed, ``false`` otherwise.
        """
        return not self.done()

    def exception(self):
        """
        Throws exception.
        :return: (Exception), exception of this Future.
        """
        self._reactor_check()
        self._event.wait()
        return self._exception

    def traceback(self):
        """
        Traceback function for the Future.
        """
        self._reactor_check()
        self._event.wait()
        return self._traceback

    def add_done_callback(self, callback):
        run_callback = False
        with self._event.condition:
            if self.done():
                run_callback = True
            else:
                self._callbacks.append(callback)

        if run_callback:
            self._invoke_cb(callback)

    def _invoke_callbacks(self):
        for callback in self._callbacks:
            self._invoke_cb(callback)

    def _invoke_cb(self, callback):
        try:
            callback(self)
        except:
            self.logger.exception("Exception when invoking callback")

    def continue_with(self, continuation_func, *args):
        """
        Create a continuation that executes when the Future is completed.

        :param continuation_func: A function which takes the future as the only parameter. Return value of the function
        will be set as the result of the continuation future.
        :return: A new Future which will be completed when the continuation is done.
        """
        future = Future()

        def callback(f):
            try:
                future.set_result(continuation_func(f, *args))
            except:
                future.set_exception(sys.exc_info()[1], sys.exc_info()[2])

        self.add_done_callback(callback)
        return future


class _Event(object):
    _flag = False

    def __init__(self):
        self.condition = threading.Condition(threading.Lock())

    def set(self):
        with self.condition:
            self._flag = True
            self.condition.notify_all()

    def is_set(self):
        return self._flag

    def wait(self):
        with self.condition:
            if not self._flag:
                self.condition.wait()
            return self._flag


class ImmediateFuture(Future):
    def __init__(self, result):
        self._result = result

    def set_exception(self, exception):
        raise NotImplementedError()

    def set_result(self, result):
        raise NotImplementedError()

    def done(self):
        return True

    def is_success(self):
        return True

    def exception(self):
        return None

    def traceback(self):
        return None

    def result(self):
        return self._result

    def add_done_callback(self, callback):
        self._invoke_cb(callback)


class ImmediateExceptionFuture(Future):
    def __init__(self, exception, traceback=None):
        self._exception = exception
        self._traceback = traceback

    def set_exception(self, exception, traceback=None):
        raise NotImplementedError()

    def set_result(self, result):
        raise NotImplementedError()

    def done(self):
        return True

    def is_success(self):
        return False

    def exception(self):
        return self._exception

    def traceback(self):
        return self._traceback

    def result(self):
        six.reraise(self._exception.__class__, self._exception, self._traceback)

    def add_done_callback(self, callback):
        self._invoke_cb(callback)


def combine_futures(*futures):
    """
    Combines set of Futures.

    :param futures: (Futures), Futures to be combined.
    :return: Result of the combination.
    """
    expected = len(futures)
    results = []
    completed = AtomicInteger()
    combined = Future()

    def done(f):
        if not combined.done():
            if f.is_success():  # TODO: ensure ordering of results as original list
                results.append(f.result())
                if completed.get_and_increment() + 1 == expected:
                    combined.set_result(results)
            else:
                combined.set_exception(f.exception(), f.traceback())

    for future in futures:
        future.add_done_callback(done)

    return combined


class _BlockingWrapper(object):
    def __init__(self, wrapped):
        self._wrapped = wrapped

    def __getattr__(self, item):
        inner = getattr(self._wrapped, item)
        if callable(inner):
            return self.wrap(inner)
        return inner

    def wrap(self, inner):
        def f(*args, **kwargs):
            result = inner(*args, **kwargs)
            if isinstance(result, Future):
                return result.result()
            return result

        return f

    def __repr__(self):
        return self._wrapped.__repr__()


def make_blocking(instance):
    """
    Takes an instance and returns an object whose methods which return non-blocking Future become blocking calls.
    :param instance: (object), a non-blocking instance.
    :return: (object), blocking version of given non-blocking instance.
    """
    return _BlockingWrapper(instance)



from hazelcast.protocol.codec import \
    queue_add_all_codec, \
    queue_add_listener_codec, \
    queue_clear_codec, \
    queue_compare_and_remove_all_codec, \
    queue_compare_and_retain_all_codec, \
    queue_contains_all_codec, \
    queue_contains_codec, \
    queue_drain_to_max_size_codec, \
    queue_is_empty_codec, \
    queue_iterator_codec, \
    queue_offer_codec, \
    queue_peek_codec, \
    queue_poll_codec, \
    queue_put_codec, \
    queue_remaining_capacity_codec, \
    queue_remove_codec, \
    queue_remove_listener_codec, \
    queue_size_codec, \
    queue_take_codec
from hazelcast.proxy.base import PartitionSpecificProxy, ItemEvent, ItemEventType
from hazelcast.util import check_not_none, to_millis


class Empty(Exception):
    pass


class Full(Exception):
    pass


class Queue(PartitionSpecificProxy):
    """
    Concurrent, blocking, distributed, observable queue. Queue is not a partitioned data-structure. All of the Queue
    content is stored in a single machine (and in the backup). Queue will not scale by adding more members
    in the cluster.
    """
    def add(self, item):
        """
        Adds the specified item to this queue if there is available space.

        :param item: (object), the specified item.
        :return: (bool), ``true`` if element is successfully added, ``false`` otherwise.
        """
        def result_fnc(f):
            if f.result():
                return True
            raise Full("Queue is full!")

        return self.offer(item).continue_with(result_fnc)

    def add_all(self, items):
        """
        Adds the elements in the specified collection to this queue.

        :param items: (Collection), collection which includes the items to be added.
        :return: (bool), ``true`` if this queue is changed after call, ``false`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))
        return self._encode_invoke(queue_add_all_codec, data_list=data_items)

    def add_listener(self, include_value=False, item_added_func=None, item_removed_func=None):
        """
        Adds an item listener for this queue. Listener will be notified for all queue add/remove events.

        :param include_value: (bool), whether received events include the updated item or not (optional).
        :param item_added_func: Function to be called when an item is added to this set (optional).
        :param item_removed_func: Function to be called when an item is deleted from this set (optional).
        :return: (str), a registration id which is used as a key to remove the listener.
        """
        request = queue_add_listener_codec.encode_request(self.name, include_value, self._is_smart)

        def handle_event_item(item, uuid, event_type):
            item = item if include_value else None
            member = self._client.cluster.get_member_by_uuid(uuid)

            item_event = ItemEvent(self.name, item, event_type, member, self._to_object)
            if event_type == ItemEventType.added:
                if item_added_func:
                    item_added_func(item_event)
            else:
                if item_removed_func:
                    item_removed_func(item_event)

        return self._register_listener(request, lambda r: queue_add_listener_codec.decode_response(r)['response'],
                                       lambda reg_id: queue_remove_listener_codec.encode_request(self.name, reg_id),
                                       lambda m: queue_add_listener_codec.handle(m, handle_event_item))

    def clear(self):
        """
        Clears this queue. Queue will be empty after this call.
        """
        return self._encode_invoke(queue_clear_codec)

    def contains(self, item):
        """
        Determines whether this queue contains the specified item or not.

        :param item: (object), the specified item to be searched.
        :return: (bool), ``true`` if the specified item exists in this queue, ``false`` otherwise.
        """
        check_not_none(item, "Item can't be None")
        item_data = self._to_data(item)
        return self._encode_invoke(queue_contains_codec, value=item_data)

    def contains_all(self, items):
        """
        Determines whether this queue contains all of the items in the specified collection or not.

        :param items: (Collection), the specified collection which includes the items to be searched.
        :return: (bool), ``true`` if all of the items in the specified collection exist in this queue, ``false`` otherwise.
        """
        check_not_none(items, "Items can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "item can't be None")
            data_items.append(self._to_data(item))
        return self._encode_invoke(queue_contains_all_codec, data_list=data_items)

    def drain_to(self, list, max_size=-1):
        """
        Transfers all available items to the given `list`_ and removes these items from this queue. If a max_size is
        specified, it transfers at most the given number of items. In case of a failure, an item can exist in both
        collections or none of them.

        This operation may be more efficient than polling elements repeatedly and putting into collection.

        :param list: (`list`_), the list where the items in this queue will be transferred.
        :param max_size: (int), the maximum number items to transfer (optional).
        :return: (int), number of transferred items.

        .. _list: https://docs.python.org/2/library/functions.html#list
        """
        def drain_result(f):
            resp = f.result()
            list.extend(resp)
            return len(resp)

        return self._encode_invoke(queue_drain_to_max_size_codec, max_size=max_size).continue_with(
            drain_result)

    def iterator(self):
        """
        Returns all of the items in this queue.

        :return: (Sequence), collection of items in this queue.
        """
        return self._encode_invoke(queue_iterator_codec)

    def is_empty(self):
        """
        Determines whether this set is empty or not.

        :return: (bool), ``true`` if this queue is empty, ``false`` otherwise.
        """
        return self._encode_invoke(queue_is_empty_codec)

    def offer(self, item, timeout=0):
        """
        Inserts the specified element into this queue if it is possible to do so immediately without violating capacity
        restrictions. Returns ``true`` upon success. If there is no space currently available:
            * If a timeout is provided, it waits until this timeout elapses and returns the result.
            * If a timeout is not provided, returns ``false`` immediately.

        :param item: (object), the item to be added.
        :param timeout: (long), maximum time in seconds to wait for addition (optional).
        :return: (bool), ``true`` if the element was added to this queue, ``false`` otherwise.
        """
        check_not_none(item, "Value can't be None")
        element_data = self._to_data(item)
        return self._encode_invoke(queue_offer_codec, value=element_data, timeout_millis=to_millis(timeout))

    def peek(self):
        """
        Retrieves the head of queue without removing it from the queue. If the queue is empty, returns ``None``.

        :return: (object), the head of this queue, or ``None`` if this queue is empty.
        """
        return self._encode_invoke(queue_peek_codec)

    def poll(self, timeout=0):
        """
        Retrieves and removes the head of this queue, if this queue is empty:
            * If a timeout is provided, it waits until this timeout elapses and returns the result.
            * If a timeout is not provided, returns ``None``.

        :param timeout: (long), maximum time in seconds to wait for addition (optional).
        :return: (object), the head of this queue, or ``None`` if this queue is empty or specified timeout elapses before an
        item is added to the queue.
        """
        return self._encode_invoke(queue_poll_codec, timeout_millis=to_millis(timeout))

    def put(self, item):
        """
        Adds the specified element into this queue. If there is no space, it waits until necessary space becomes
        available.

        :param item: (object), the specified item.
        """
        check_not_none(item, "Value can't be None")
        element_data = self._to_data(item)
        return self._encode_invoke(queue_put_codec, value=element_data)

    def remaining_capacity(self):
        """
        Returns the remaining capacity of this queue.

        :return: (int), remaining capacity of this queue.
        """
        return self._encode_invoke(queue_remaining_capacity_codec)

    def remove(self, item):
        """
        Removes the specified element from the queue if it exists.

        :param item: (object), the specified element to be removed.
        :return: (bool), ``true`` if the specified element exists in this queue.
        """
        check_not_none(item, "Value can't be None")
        item_data = self._to_data(item)
        return self._encode_invoke(queue_remove_codec, value=item_data)

    def remove_all(self, items):
        """
        Removes all of the elements of the specified collection from this queue.

        :param items: (Collection), the specified collection.
        :return: (bool), ``true`` if the call changed this queue, ``false`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))
        return self._encode_invoke(queue_compare_and_remove_all_codec, data_list=data_items)

    def remove_listener(self, registration_id):
        """
        Removes the specified item listener. Returns silently if the specified listener was not added before.

        :param registration_id: (str), id of the listener to be deleted.
        :return: (bool), ``true`` if the item listener is removed, ``false`` otherwise.
        """
        return self._deregister_listener(registration_id)

    def retain_all(self, items):
        """
        Removes the items which are not contained in the specified collection. In other words, only the items that
        are contained in the specified collection will be retained.

        :param items: (Collection), collection which includes the elements to be retained in this set.
        :return: (bool), ``true`` if this queue changed as a result of the call.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))
        return self._encode_invoke(queue_compare_and_retain_all_codec, data_list=data_items)

    def size(self):
        """
        Returns the number of elements in this collection. If the size is greater than sys.maxint, it returns
        sys.maxint.

        :return: (int), size of the queue.
        """
        return self._encode_invoke(queue_size_codec)

    def take(self):
        """
        Retrieves and removes the head of this queue, if necessary, waits until an item becomes available.

        :return: (object), the head of this queue.
        """

        return self._encode_invoke(queue_take_codec)



from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import ClientMessage
from hazelcast.protocol.codec.flake_id_generator_message_type import *

REQUEST_TYPE = FLAKEIDGENERATOR_NEWIDBATCH
RESPONSE_TYPE = 126
RETRYABLE = True


def calculate_size(name, batch_size):
    """ Calculates the request payload size"""
    data_size = 0
    data_size += calculate_size_str(name)
    data_size += INT_SIZE_IN_BYTES
    return data_size


def encode_request(name, batch_size):
    """ Encode request into client_message"""
    client_message = ClientMessage(payload_size=calculate_size(name, batch_size))
    client_message.set_message_type(REQUEST_TYPE)
    client_message.set_retryable(RETRYABLE)
    client_message.append_str(name)
    client_message.append_int(batch_size)
    client_message.update_frame_length()
    return client_message


def decode_response(client_message, to_object=None):
    """ Decode response from client message"""
    parameters = dict(base=None, increment=None, batch_size=None)
    parameters['base'] = client_message.read_long()
    parameters['increment'] = client_message.read_long()
    parameters['batch_size'] = client_message.read_int()
    return parameters







LIST_SIZE = 0x0501
LIST_CONTAINS = 0x0502
LIST_CONTAINSALL = 0x0503
LIST_ADD = 0x0504
LIST_REMOVE = 0x0505
LIST_ADDALL = 0x0506
LIST_COMPAREANDREMOVEALL = 0x0507
LIST_COMPAREANDRETAINALL = 0x0508
LIST_CLEAR = 0x0509
LIST_GETALL = 0x050a
LIST_ADDLISTENER = 0x050b
LIST_REMOVELISTENER = 0x050c
LIST_ISEMPTY = 0x050d
LIST_ADDALLWITHINDEX = 0x050e
LIST_GET = 0x050f
LIST_SET = 0x0510
LIST_ADDWITHINDEX = 0x0511
LIST_REMOVEWITHINDEX = 0x0512
LIST_LASTINDEXOF = 0x0513
LIST_INDEXOF = 0x0514
LIST_SUB = 0x0515
LIST_ITERATOR = 0x0516
LIST_LISTITERATOR = 0x0517




RINGBUFFER_SIZE = 0x1901
RINGBUFFER_TAILSEQUENCE = 0x1902
RINGBUFFER_HEADSEQUENCE = 0x1903
RINGBUFFER_CAPACITY = 0x1904
RINGBUFFER_REMAININGCAPACITY = 0x1905
RINGBUFFER_ADD = 0x1906
RINGBUFFER_READONE = 0x1908
RINGBUFFER_ADDALL = 0x1909
RINGBUFFER_READMANY = 0x190a
